import pytest
import pymongo

import datetime
import uuid
import itertools

from fedn.network.storage.dbconnection import DatabaseConnection

@pytest.fixture
def test_validations():
    start_time = datetime.datetime(2021, 1, 4, 1, 2, 4)
   
    return [{"data":"test_data7", "timestamp":start_time+datetime.timedelta(days=76), "correlation_id":str(uuid.uuid4())},
            {"data":"test_data2", "timestamp":start_time+datetime.timedelta(days=22), "correlation_id":str(uuid.uuid4())},
            {"data":"test_data4", "timestamp":start_time+datetime.timedelta(days=1452), "correlation_id":str(uuid.uuid4())},
            {"data":"test_data5", "timestamp":start_time+datetime.timedelta(days=52), "correlation_id":str(uuid.uuid4())},
            {"data":"test_data1", "timestamp":start_time+datetime.timedelta(days=212), "correlation_id":str(uuid.uuid4())},
            {"data":"test_data3", "timestamp":start_time+datetime.timedelta(days=12), "correlation_id":str(uuid.uuid4())},
            {"data":"test_data8", "timestamp":start_time+datetime.timedelta(days=72), "correlation_id":str(uuid.uuid4())},
            {"data":"test_data6", "timestamp":start_time+datetime.timedelta(days=42), "correlation_id":str(uuid.uuid4())}]

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_validations):
    for c in test_validations:
        res, _ = mongo_connection.validation_store.add(c)
        assert res == True

    for c in test_validations:
        res, _ = postgres_connection.validation_store.add(c)
        assert res == True

    for c in test_validations:
        res, _ = sql_connection.validation_store.add(c)
        assert res == True

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    # TODO:Clean up



@pytest.fixture
def options():
    sorting_keys = (#None, 
                        "data",
                        "timestamp",
                        #"invalid_key"
                        ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"data":"test_data6"}, {"data":""})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestValidationStore:

    def test_add_update_delete(self, postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection):
        pass

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                res = db_1.validation_store.list(*opt, **kwargs)
                count, gathered_validations = res["count"], res["result"]

                res = db_2.validation_store.list(*opt, **kwargs)
                count2, gathered_validations2 = res["count"], res["result"]
                #TODO: The count is not equal to the number of clients in the list, but the number of clients returned by the query before skip and limit
                #It is not clear what is the intended behavior
                # assert(count == len(gathered_clients))
                # assert count == count2
                assert len(gathered_validations) == len(gathered_validations2)

                for i in range(len(gathered_validations)):
                    assert gathered_validations2[i]["data"] == gathered_validations[i]["data"]
                    #TODO: timestamp is not equal between the two databases, different formats
                    #assert gathered_validations2[i]["timestamp"] == gathered_validations[i]["timestamp"]
                    assert gathered_validations2[i]["correlation_id"] == gathered_validations[i]["correlation_id"]






from typing import Tuple

import pymongo

api_version = "v1"


def is_positive_integer(s):
    return s is not None and s.isdigit() and int(s) > 0


def get_limit(headers: object) -> int:
    limit: str = headers.get("X-Limit")
    if is_positive_integer(limit):
        return int(limit)
    return 0


def get_reverse(headers: object) -> bool:
    reverse: str = headers.get("X-Reverse")
    if reverse and reverse.lower() == "true":
        return True
    return False


def get_skip(headers: object) -> int:
    skip: str | None = headers.get("X-Skip")
    if is_positive_integer(skip):
        return int(skip)
    return 0


def get_typed_list_headers(
    headers: object,
) -> Tuple[int, int, str, int, bool]:
    sort_key: str = headers.get("X-Sort-Key")
    sort_order: str = headers.get("X-Sort-Order")

    limit: int = get_limit(headers)
    skip: int = get_skip(headers)

    if sort_order is not None:
        sort_order = pymongo.ASCENDING if sort_order.lower() == "asc" else pymongo.DESCENDING
    else:
        sort_order = pymongo.DESCENDING

    return limit, skip, sort_key, sort_order


def get_post_data_to_kwargs(request: object) -> dict:
    try:
        # Try to get data from form
        request_data = request.form.to_dict()
    except Exception:
        request_data = None

    if not request_data:
        try:
            # Try to get data from JSON
            request_data = request.get_json()
        except Exception:
            request_data = {}

    kwargs = {}
    for key, value in request_data.items():
        if isinstance(value, str) and "," in value:
            kwargs[key] = {"$in": value.split(",")}
        else:
            kwargs[key] = value

    return kwargs



"""Shared objects for the network API."""

import os
from typing import Tuple

from werkzeug.security import safe_join

from fedn.common.config import get_modelstorage_config, get_network_config
from fedn.network.controller.control import Control
from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.s3.base import RepositoryBase
from fedn.network.storage.s3.miniorepository import MINIORepository
from fedn.network.storage.s3.repository import Repository
from fedn.network.storage.s3.saasrepository import SAASRepository
from fedn.network.storage.statestore.stores.analytic_store import AnalyticStore
from fedn.network.storage.statestore.stores.client_store import ClientStore
from fedn.network.storage.statestore.stores.combiner_store import CombinerStore
from fedn.network.storage.statestore.stores.model_store import ModelStore
from fedn.network.storage.statestore.stores.package_store import PackageStore
from fedn.network.storage.statestore.stores.prediction_store import PredictionStore
from fedn.network.storage.statestore.stores.round_store import RoundStore
from fedn.network.storage.statestore.stores.session_store import SessionStore
from fedn.network.storage.statestore.stores.status_store import StatusStore
from fedn.network.storage.statestore.stores.validation_store import ValidationStore
from fedn.utils.checksum import sha

modelstorage_config = get_modelstorage_config()
network_id = get_network_config()

# TODO: Refactor all access to the stores to use the DatabaseConnection
stores = DatabaseConnection().get_stores()
session_store: SessionStore = stores.session_store
model_store: ModelStore = stores.model_store
round_store: RoundStore = stores.round_store
package_store: PackageStore = stores.package_store
combiner_store: CombinerStore = stores.combiner_store
client_store: ClientStore = stores.client_store
status_store: StatusStore = stores.status_store
validation_store: ValidationStore = stores.validation_store
prediction_store: PredictionStore = stores.prediction_store
analytic_store: AnalyticStore = stores.analytic_store


repository = Repository(modelstorage_config["storage_config"], storage_type=modelstorage_config["storage_type"])

control = Control(
    network_id=network_id,
    session_store=session_store,
    model_store=model_store,
    round_store=round_store,
    package_store=package_store,
    combiner_store=combiner_store,
    client_store=client_store,
    model_repository=repository,
)

# TODO: use Repository
minio_repository: RepositoryBase = None

storage_type = os.environ.get("FEDN_STORAGE_TYPE", modelstorage_config["storage_type"])
if storage_type == "MINIO":
    minio_repository = MINIORepository(modelstorage_config["storage_config"])
elif storage_type == "SAAS":
    minio_repository = SAASRepository(modelstorage_config["storage_config"])
else:
    minio_repository = MINIORepository(modelstorage_config["storage_config"])


def get_checksum(name: str = None) -> Tuple[bool, str, str]:
    """Generate a checksum for a given file."""
    message = None
    sum = None
    success = False

    if name is None:
        active_package = package_store.get_active()
        if active_package is None:
            message = "No compute package uploaded"
            return success, message, sum
        name = active_package["storage_file_name"]
    file_path = safe_join(os.getcwd(), name)
    try:
        sum = str(sha(file_path))
        success = True
        message = "Checksum created."
    except FileNotFoundError:
        message = "File not found."
    return success, message, sum



from fedn.common.config import get_modelstorage_config
from fedn.network.combiner.modelservice import ModelService
from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.s3.repository import Repository
from fedn.network.storage.statestore.stores.client_store import ClientStore
from fedn.network.storage.statestore.stores.combiner_store import CombinerStore
from fedn.network.storage.statestore.stores.model_store import ModelStore
from fedn.network.storage.statestore.stores.package_store import PackageStore
from fedn.network.storage.statestore.stores.prediction_store import PredictionStore
from fedn.network.storage.statestore.stores.round_store import RoundStore
from fedn.network.storage.statestore.stores.session_store import SessionStore
from fedn.network.storage.statestore.stores.status_store import StatusStore
from fedn.network.storage.statestore.stores.validation_store import ValidationStore

modelstorage_config = get_modelstorage_config()

# TODO: Refactor all access to the stores to use the DatabaseConnection
stores = DatabaseConnection().get_stores()
session_store: SessionStore = stores.session_store
model_store: ModelStore = stores.model_store
round_store: RoundStore = stores.round_store
package_store: PackageStore = stores.package_store
combiner_store: CombinerStore = stores.combiner_store
client_store: ClientStore = stores.client_store
status_store: StatusStore = stores.status_store
validation_store: ValidationStore = stores.validation_store
prediction_store: PredictionStore = stores.prediction_store


repository = Repository(modelstorage_config["storage_config"], init_buckets=False)

modelservice = ModelService()



import datetime
import uuid
from typing import Any, Dict, List, Tuple

import pymongo
from bson import ObjectId
from pymongo.database import Database
from sqlalchemy import func, select

from fedn.network.storage.statestore.stores.shared import from_document
from fedn.network.storage.statestore.stores.sql.shared import SessionConfigModel, SessionModel
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store


class SessionConfig:
    def __init__(
        self,
        aggregator: str,
        round_timeout: int,
        buffer_size: int,
        model_id: str,
        delete_models_storage: bool,
        clients_required: int,
        validate: bool,
        helper_type: str,
    ):
        self.aggregator = aggregator
        self.round_timeout = round_timeout
        self.buffer_size = buffer_size
        self.model_id = model_id
        self.delete_models_storage = delete_models_storage
        self.clients_required = clients_required
        self.validate = validate
        self.helper_type = helper_type


class Session:
    def __init__(self, id: str, session_id: str, status: str, committed_at: datetime, name: str = None, session_config: SessionConfig = None):
        self.id = id
        self.session_id = session_id
        self.status = status
        self.committed_at = committed_at
        self.session_config = session_config
        self.name = name


class SessionStore(Store[Session]):
    pass


def validate_session_config(session_config: SessionConfig) -> Tuple[bool, str]:
    if "aggregator" not in session_config:
        return False, "session_config.aggregator is required"

    if "round_timeout" not in session_config:
        return False, "session_config.round_timeout is required"

    if not isinstance(session_config["round_timeout"], (int, float)):
        return False, "session_config.round_timeout must be an integer"

    if "buffer_size" not in session_config:
        return False, "session_config.buffer_size is required"

    if not isinstance(session_config["buffer_size"], int):
        return False, "session_config.buffer_size must be an integer"

    if "model_id" not in session_config or session_config["model_id"] == "":
        return False, "session_config.model_id is required"

    if not isinstance(session_config["model_id"], str):
        return False, "session_config.model_id must be a string"

    if "delete_models_storage" not in session_config:
        return False, "session_config.delete_models_storage is required"

    if not isinstance(session_config["delete_models_storage"], bool):
        return False, "session_config.delete_models_storage must be a boolean"

    if "clients_required" not in session_config:
        return False, "session_config.clients_required is required"

    if not isinstance(session_config["clients_required"], int):
        return False, "session_config.clients_required must be an integer"

    if "validate" not in session_config:
        return False, "session_config.validate is required"

    if not isinstance(session_config["validate"], bool):
        return False, "session_config.validate must be a boolean"

    if "helper_type" not in session_config or session_config["helper_type"] == "":
        return False, "session_config.helper_type is required"

    if not isinstance(session_config["helper_type"], str):
        return False, "session_config.helper_type must be a string"

    return True, ""


def validate(item: Session) -> Tuple[bool, str]:
    if "session_config" not in item or item["session_config"] is None:
        return False, "session_config is required"

    session_config = None

    if isinstance(item["session_config"], dict):
        session_config = item["session_config"]
    elif isinstance(item["session_config"], list):
        session_config = item["session_config"][0]
    else:
        return False, "session_config must be a dict"

    return validate_session_config(session_config)


def complement(item: Session):
    item["status"] = "Created"
    item["committed_at"] = datetime.datetime.now()

    if "session_id" not in item or item["session_id"] == "" or not isinstance(item["session_id"], str):
        item["session_id"] = str(uuid.uuid4())


class MongoDBSessionStore(MongoDBStore[Session]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)
        self.database[self.collection].create_index([("session_id", pymongo.DESCENDING)])

    def get(self, id: str) -> Session:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the session_id (property)
        return: The entity
        """
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            document = self.database[self.collection].find_one({"_id": id_obj})
        else:
            document = self.database[self.collection].find_one({"session_id": id})

        if document is None:
            return None

        return from_document(document)

    def update(self, id: str, item: Session) -> Tuple[bool, Any]:
        valid, message = validate(item)
        if not valid:
            return False, message

        return super().update(id, item)

    def add(self, item: Session) -> Tuple[bool, Any]:
        """Add an entity
        param item: The entity to add
            type: Session
            description: The entity to add
        return: A tuple with a boolean indicating success and the entity
        """
        valid, message = validate(item)
        if not valid:
            return False, message

        complement(item)

        return super().add(item)

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for SessionStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs) -> Dict[int, List[Session]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
            description: The maximum number of entities to return
        param skip: The number of entities to skip
            type: int
            description: The number of entities to skip
        param sort_key: The key to sort by
            type: str
            description: The key to sort by
        param sort_order: The order to sort by
            type: pymongo.DESCENDING
            description: The order to sort by
        param kwargs: Additional query parameters
            type: dict
            description: Additional query parameters
        return: The entities
        """
        return super().list(limit, skip, sort_key or "session_id", sort_order, **kwargs)


def from_row(row: dict) -> Session:
    return {
        "id": row["id"],
        "name": row["name"],
        "session_id": row["id"],
        "status": row["status"],
        "committed_at": row["committed_at"],
        "session_config": {
            "aggregator": row["aggregator"],
            "round_timeout": row["round_timeout"],
            "buffer_size": row["buffer_size"],
            "model_id": row["model_id"],
            "delete_models_storage": row["delete_models_storage"],
            "clients_required": row["clients_required"],
            "validate": row["validate"],
            "helper_type": row["helper_type"],
        },
    }


class SQLSessionStore(SessionStore, SQLStore[Session]):
    def __init__(self, Session):
        super().__init__(Session)

    def get(self, id: str) -> Session:
        with self.Session() as session:
            stmt = select(SessionModel, SessionConfigModel).join(SessionModel.session_config).where(SessionModel.id == id)
            item = session.execute(stmt).first()
            if item is None:
                return None
            s, c = item
            combined_dict = {
                "id": s.id,
                "name": s.name,
                "session_id": s.id,
                "status": s.status,
                "committed_at": s.committed_at,
                "aggregator": c.aggregator,
                "round_timeout": c.round_timeout,
                "buffer_size": c.buffer_size,
                "model_id": c.model_id,
                "delete_models_storage": c.delete_models_storage,
                "clients_required": c.clients_required,
                "validate": c.validate,
                "helper_type": c.helper_type,
            }
            return from_row(combined_dict)

    def update(self, id: str, item: Session) -> Tuple[bool, Any]:
        valid, message = validate(item)
        if not valid:
            return False, message
        with self.Session() as session:
            stmt = select(SessionModel, SessionConfigModel).join(SessionModel.session_config).where(SessionModel.id == id)
            existing_item = session.execute(stmt).first()
            if existing_item is None:
                return False, f"Entity not found {id}"
            s, c = existing_item

            s.name = item["name"] if "name" in item else None
            s.status = item["status"]
            s.committed_at = item["committed_at"]

            session_config = item["session_config"]

            c.aggregator = session_config["aggregator"]
            c.round_timeout = session_config["round_timeout"]
            c.buffer_size = session_config["buffer_size"]
            c.model_id = session_config["model_id"]
            c.delete_models_storage = session_config["delete_models_storage"]
            c.clients_required = session_config["clients_required"]
            c.validate = session_config["validate"]
            c.helper_type = session_config["helper_type"]

            session.commit()

            combined_dict = {
                "id": s.id,
                "name": s.name,
                "session_id": s.id,
                "status": s.status,
                "committed_at": s.committed_at,
                "aggregator": c.aggregator,
                "round_timeout": c.round_timeout,
                "buffer_size": c.buffer_size,
                "model_id": c.model_id,
                "delete_models_storage": c.delete_models_storage,
                "clients_required": c.clients_required,
                "validate": c.validate,
                "helper_type": c.helper_type,
            }

            return True, from_row(combined_dict)

    def add(self, item: Session) -> Tuple[bool, Any]:
        valid, message = validate(item)
        if not valid:
            return False, message

        complement(item)

        with self.Session() as session:
            parent_item = SessionModel(
                id=item["session_id"], status=item["status"], name=item["name"] if "name" in item else None, committed_at=item["committed_at"] or None
            )

            session_config = item["session_config"]

            child_item = SessionConfigModel(
                aggregator=session_config["aggregator"],
                round_timeout=session_config["round_timeout"],
                buffer_size=session_config["buffer_size"],
                model_id=session_config["model_id"],
                delete_models_storage=session_config["delete_models_storage"],
                clients_required=session_config["clients_required"],
                validate=session_config["validate"],
                helper_type=session_config["helper_type"],
            )
            child_item.session = parent_item

            session.add(child_item)
            session.add(parent_item)
            session.commit()

            combined_dict = {
                "id": parent_item.id,
                "name": parent_item.name,
                "session_id": parent_item.id,
                "status": parent_item.status,
                "committed_at": parent_item.committed_at,
                "aggregator": child_item.aggregator,
                "round_timeout": child_item.round_timeout,
                "buffer_size": child_item.buffer_size,
                "model_id": child_item.model_id,
                "delete_models_storage": child_item.delete_models_storage,
                "clients_required": child_item.clients_required,
                "validate": child_item.validate,
                "helper_type": child_item.helper_type,
            }

            return True, from_row(combined_dict)

    def delete(self, id: str) -> bool:
        raise NotImplementedError

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs):
        with self.Session() as session:
            stmt = select(SessionModel, SessionConfigModel).join(SessionModel.session_config)
            for key, value in kwargs.items():
                if "session_config" in key:
                    key = key.replace("session_config.", "")
                    stmt = stmt.where(getattr(SessionConfigModel, key) == value)
                else:
                    stmt = stmt.where(getattr(SessionModel, key) == value)

            if sort_key:
                _sort_order: str = "DESC" if sort_order == pymongo.DESCENDING else "ASC"
                _sort_key: str = sort_key or "committed_at"

                if _sort_key in SessionModel.__table__.columns:
                    sort_obj = SessionModel.__table__.columns.get(_sort_key) if _sort_order == "ASC" else SessionModel.__table__.columns.get(_sort_key).desc()

                    stmt = stmt.order_by(sort_obj)

            if limit:
                stmt = stmt.offset(skip or 0).limit(limit)
            elif skip:
                stmt = stmt.offset(skip)

            items = session.execute(stmt)

            result = []

            for item in items:
                s, c = item
                combined_dict = {
                    "id": s.id,
                    "session_id": s.id,
                    "name": s.name,
                    "status": s.status,
                    "committed_at": s.committed_at,
                    "aggregator": c.aggregator,
                    "round_timeout": c.round_timeout,
                    "buffer_size": c.buffer_size,
                    "model_id": c.model_id,
                    "delete_models_storage": c.delete_models_storage,
                    "clients_required": c.clients_required,
                    "validate": c.validate,
                    "helper_type": c.helper_type,
                }
                result.append(from_row(combined_dict))

            return {"count": len(result), "result": result}

    def count(self, **kwargs):
        with self.Session() as session:
            stmt = select(func.count()).select_from(SessionModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(SessionModel, key) == value)

            count = session.scalar(stmt)

            return count



import os
from getpass import getpass

import click
import requests

from .main import main
from .shared import STUDIO_DEFAULTS, get_response, set_context

# Replace this with the platform's actual login endpoint
home_dir = os.path.expanduser("~")


@main.group("studio")
@click.pass_context
def login_cmd(ctx):
    """:param ctx:"""
    pass


@login_cmd.command("login")
@click.option("-u", "--username", required=False, default=None, help="username in studio")
@click.option("-P", "--password", required=False, default=None, help="password in studio")
@click.option("-p", "--protocol", required=False, default=STUDIO_DEFAULTS["protocol"], help="Communication protocol of studio (api)")
@click.option("-H", "--host", required=False, default=STUDIO_DEFAULTS["host"], help="Hostname of studio (api)")
@click.pass_context
def login_cmd(ctx, protocol: str, host: str, username: str, password: str):
    """Login to FEDn Studio"""
    # Step 1: Display welcome message
    click.secho("Welcome to Scaleout FEDn!", fg="green")

    url = f"{protocol}://{host}/api/token/"

    # Step 3: Prompt for username and password
    if username is None:
        username = input("Please enter your username: ")
    if password is None:
        password = getpass("Please enter your password: ")

    # Call the authentication API
    try:
        response = requests.post(url, json={"username": username, "password": password}, headers={"Content-Type": "application/json"})
        response.raise_for_status()  # Raise an error for HTTP codes 4xx/5xx
    except requests.exceptions.RequestException as e:
        click.secho("Error connecting to the platform. Please try again.", fg="red")
        click.secho(str(e), fg="red")
        return

    # Handle the response
    if response.status_code == 200:
        context_data = get_context(response, protocol, host)

        context_path = os.path.join(home_dir, ".fedn")
        if not os.path.exists(context_path):
            os.makedirs(context_path)
        set_context(context_path, context_data)
    else:
        click.secho(f"Unexpected error: {response.status_code}", fg="red")


# Sets the context for a given user
def get_context(response, protocol: str, host: str):
    """Generates content for context file with the following data:
    User tokens: access and refresh token to authenticate user towards Studio
    Active project tokens: access and refresh token to authenticate user towards controller
    Active project id: slug of active project
    Active project url: controller url of active project
    """
    context_data = {"User tokens": {}, "Active project tokens": {}, "Active project id": {}, "Active project url": {}}
    user_token_data = response.json()
    if user_token_data.get("access"):
        context_data["User tokens"] = user_token_data
        studio_api = True
        headers_projects = {}
        user_access_token = user_token_data.get("access")
        response_projects = get_response(
            protocol=protocol,
            host=host,
            port=None,
            endpoint="projects",
            token=user_access_token,
            headers=headers_projects,
            usr_api=studio_api,
            usr_token=True,
        )
        if response_projects.status_code == 200:
            projects_response_json = response_projects.json()
            if len(projects_response_json) > 0:
                id = projects_response_json[0].get("slug")
                context_data["Active project id"] = id
                headers_projects["X-Project-Slug"] = id
                response_project_tokens = get_response(
                    protocol=protocol,
                    host=host,
                    port=None,
                    endpoint="admin-token",
                    token=user_access_token,
                    headers=headers_projects,
                    usr_api=studio_api,
                    usr_token=False,
                )
                if response_project_tokens.status_code == 200:
                    project_tokens = response_project_tokens.json()
                    context_data["Active project tokens"] = project_tokens
                    controller_url = f"{protocol}://{host}/{id}-fedn-reducer"
                    context_data["Active project url"] = controller_url
                    click.secho("Login successful!", fg="green")
                else:
                    click.secho(f"Unexpected error: {response_project_tokens.status_code}", fg="red")
        else:
            click.secho(f"Unexpected error: {response_projects.status_code}", fg="red")
    else:
        click.secho("Login failed. Please check your credentials.", fg="red")

    return context_data



# Unittest for Flask API endpoints
#
# Run with:
# python -m unittest fedn.tests.network.api.tests
#
# or
#
# python3 -m unittest fedn.tests.network.api.tests
#
# or
#
# python3 -m unittest fedn.tests.network.api.tests.NetworkAPITests
#
# or
#
# python -m unittest fedn.tests.network.api.tests.NetworkAPITests
#
# or
#
# python -m unittest fedn.tests.network.api.tests.NetworkAPITests.test_get_model_trail
#
# or
#
# python3 -m unittest fedn.tests.network.api.tests.NetworkAPITests.test_get_model_trail
#

import unittest
from unittest.mock import patch, MagicMock

import pymongo

import fedn  # noqa: F401

entitites = ['clients', 'combiners', 'models', 'packages', 'rounds', 'sessions', 'statuses', 'validations']

class NetworkAPITests(unittest.TestCase):
    """ Unittests for the Network API. """
    @patch('fedn.network.controller.controlbase.ControlBase', autospec=True)
    def setUp(self, mock_control):
        # start Flask server in testing mode
        import fedn.network.api.server
        self.app = fedn.network.api.server.app.test_client()


    def test_health(self):
        """ Test get_models endpoint. """
        response = self.app.get('/health')
        # Assert response
        self.assertEqual(response.status_code, 200)

    def test_add_combiner(self):
        """ Test get_models endpoint. """
        # Mock api.get_models
        
        response = self.app.post('/add_combiner')
        
        self.assertEqual(response.status_code, 410)

    def test_get_controller_status(self):
        """ Test get_models endpoint. """
        response = self.app.get('/get_controller_status')
        # Assert response
        self.assertEqual(response.status_code, 200)

    def test_get_endpoints(self):
        """ Test allt get endpoints. """
        return_value = {"count": 1, "results": [{"id": "test"}]}
        fedn.network.api.shared.client_store.list = MagicMock(return_value=return_value)
        fedn.network.api.shared.combiner_store.list = MagicMock(return_value=return_value)
        fedn.network.api.shared.model_store.list = MagicMock(return_value=return_value)
        fedn.network.api.shared.package_store.list = MagicMock(return_value=return_value)
        fedn.network.api.shared.round_store.list = MagicMock(return_value=return_value)
        fedn.network.api.shared.session_store.list = MagicMock(return_value=return_value)
        fedn.network.api.shared.status_store.list = MagicMock(return_value=return_value)
        fedn.network.api.shared.validation_store.list = MagicMock(return_value=return_value)

        for entity in entitites:
            response = self.app.get(f'/api/v1/{entity}/')
            # Assert response
            self.assertEqual(response.status_code, 200)

            count = response.json['count']
            expected_count = return_value['count']

            self.assertEqual(count, expected_count)

            id = response.json['results'][0]['id']
            expected_id = return_value['results'][0]['id']

            self.assertEqual(id, expected_id)

        fedn.network.api.shared.client_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.client_store.list.assert_called_once()
        fedn.network.api.shared.combiner_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.combiner_store.list.assert_called_once()
        fedn.network.api.shared.model_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.model_store.list.assert_called_once()
        fedn.network.api.shared.package_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.package_store.list.assert_called_once()
        fedn.network.api.shared.round_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.round_store.list.assert_called_once()        
        fedn.network.api.shared.session_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.session_store.list.assert_called_once()
        fedn.network.api.shared.status_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.status_store.list.assert_called_once()
        fedn.network.api.shared.validation_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.validation_store.list.assert_called_once()
   
        for entity in entitites:
            headers = {
                "X-Limit": 10,
                "X-Skip": 10,
                "X-Sort-Key": "test",
                "X-Sort-Order": "asc"
            }
            response = self.app.get(f'/api/v1/{entity}/', headers=headers)
            # Assert response
            self.assertEqual(response.status_code, 200)

        fedn.network.api.shared.client_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.combiner_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.model_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.package_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.round_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)        
        fedn.network.api.shared.session_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.status_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.validation_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
           

if __name__ == '__main__':
    unittest.main()



import graphene
import pymongo

from fedn.network.api.shared import model_store, session_store, status_store, validation_store


class ActorType(graphene.ObjectType):
    name = graphene.String()
    role = graphene.String()


class StatusType(graphene.ObjectType):
    data = graphene.String()
    extra = graphene.String()
    id = graphene.String()
    logLevel = graphene.String()  # noqa: N815
    sender = graphene.Field(ActorType)
    sessionId = graphene.String()  # noqa: N815
    status = graphene.String()
    timestamp = graphene.String()
    type = graphene.String()


class ValidationType(graphene.ObjectType):
    correlationId = graphene.String()  # noqa: N815
    data = graphene.String()
    id = graphene.String()
    meta = graphene.String()
    modelId = graphene.String()  # noqa: N815
    receiver = graphene.Field(ActorType)
    sender = graphene.Field(ActorType)
    sessionId = graphene.String()  # noqa: N815
    timestamp = graphene.String()

    def resolve_receiver(self, info):
        return self["receiver"]

    def resolve_sender(self, info):
        return self["sender"]


class ModelType(graphene.ObjectType):
    id = graphene.String()
    model = graphene.String()
    name = graphene.String()
    committed_at = graphene.DateTime()
    session_id = graphene.String()
    parent_model = graphene.String()
    validations = graphene.List(ValidationType)

    def resolve_validations(self, info):
        kwargs = {"modelId": self["model"]}
        response = validation_store.list(0, 0, None, sort_order=pymongo.DESCENDING, **kwargs)
        result = response["result"]

        return result


class SessionConfigType(graphene.ObjectType):
    aggregator = graphene.String()
    round_timeout = graphene.Int()
    buffer_size = graphene.Int()
    model_id = graphene.String()
    delete_models_storage = graphene.Boolean()
    clients_required = graphene.Int()
    helper_type = graphene.String()
    validate = graphene.Boolean()
    session_id = graphene.String()
    model_id = graphene.String()


class SessionType(graphene.ObjectType):
    id = graphene.String()
    session_id = graphene.String()
    name = graphene.String()
    committed_at = graphene.DateTime()
    session_config = graphene.Field(SessionConfigType)
    models = graphene.List(ModelType)
    validations = graphene.List(ValidationType)
    statuses = graphene.List(StatusType)

    def resolve_session_config(self, info):
        return self["session_config"]

    def resolve_models(self, info):
        kwargs = {"session_id": self["session_id"]}
        response = model_store.list(0, 0, None, sort_order=pymongo.DESCENDING, **kwargs)
        result = response["result"]

        return result

    def resolve_validations(self, info):
        kwargs = {"sessionId": self["session_id"]}
        response = validation_store.list(0, 0, None, sort_order=pymongo.DESCENDING, **kwargs)
        result = response["result"]

        return result

    def resolve_statuses(self, info):
        kwargs = {"sessionId": self["session_id"]}
        response = status_store.list(0, 0, None, sort_order=pymongo.DESCENDING, **kwargs)
        result = response["result"]

        return result


class Query(graphene.ObjectType):
    session = graphene.Field(
        SessionType,
        id=graphene.String(required=True),
    )
    sessions = graphene.List(
        SessionType,
        name=graphene.String(),
    )

    model = graphene.Field(
        ModelType,
        id=graphene.String(required=True),
    )

    models = graphene.List(
        ModelType,
        session_id=graphene.String(),
    )

    validation = graphene.Field(
        ValidationType,
        id=graphene.String(required=True),
    )

    validations = graphene.List(
        ValidationType,
        session_id=graphene.String(),
    )

    status = graphene.Field(
        StatusType,
        id=graphene.String(required=True),
    )

    statuses = graphene.List(
        StatusType,
        session_id=graphene.String(),
    )

    def resolve_session(root, info, id: str = None):
        result = session_store.get(id)

        return result

    def resolve_sessions(root, info, name: str = None):
        response = None
        if name:
            kwargs = {"name": name}
            response = session_store.list(0, 0, None, sort_order=pymongo.DESCENDING, **kwargs)
        else:
            response = session_store.list(0, 0, None)

        return response["result"]

    def resolve_model(root, info, id: str = None):
        result = model_store.get(id)

        return result

    def resolve_models(root, info, session_id: str = None):
        response = None
        if session_id:
            kwargs = {"session_id": session_id}
            response = model_store.list(0, 0, None, sort_order=pymongo.DESCENDING, **kwargs)
        else:
            response = model_store.list(0, 0, None)

        return response["result"]

    def resolve_validation(root, info, id: str = None):
        result = validation_store.get(id)

        return result

    def resolve_validations(root, info, session_id: str = None):
        response = None
        if session_id:
            kwargs = {"session_id": session_id}
            response = validation_store.list(0, 0, None, sort_order=pymongo.DESCENDING, **kwargs)
        else:
            response = validation_store.list(0, 0, None)

        return response["result"]

    def resolve_status(root, info, id: str = None):
        result = status_store.get(id)

        return result

    def resolve_statuses(root, info, session_id: str = None):
        response = None
        if session_id:
            kwargs = {"sessionId": session_id}
            response = status_store.list(0, 0, None, sort_order=pymongo.DESCENDING, **kwargs)
        else:
            response = status_store.list(0, 0, None)

        return response["result"]


schema = graphene.Schema(query=Query)



"""Package commands for the CLI."""

import fnmatch
import os
import sys
import tarfile

import click

from fedn.cli.main import main
from fedn.cli.shared import CONTROLLER_DEFAULTS, get_response, print_response
from fedn.common.log_config import logger


def create_tar_with_ignore(path: str, name: str) -> None:
    """Create a tar archive from a directory with an ignore and fedn.yaml file."""
    try:
        ignore_patterns = []
        ignore_file = os.path.join(path, ".fednignore")
        if os.path.exists(ignore_file):
            # Read ignore patterns from .fednignore file
            with open(ignore_file, "r") as f:
                ignore_patterns = [line.strip() for line in f if line.strip() and not line.startswith("#")]

        def is_ignored(file_path: str) -> bool:
            relative_path = os.path.relpath(file_path, path)
            return any(fnmatch.fnmatch(relative_path, pattern) or fnmatch.fnmatch(os.path.basename(file_path), pattern) for pattern in ignore_patterns)

        tar_path = os.path.join(path, name)
        with tarfile.open(tar_path, "w:gz") as tar:
            for root, dirs, files in os.walk(path):
                dirs[:] = [d for d in dirs if not is_ignored(os.path.join(root, d))]
                for file in files:
                    file_path = os.path.join(root, file)
                    if not is_ignored(file_path):
                        logger.debug(f"Adding file to tar archive: {file_path}")
                        tar.add(file_path, arcname=os.path.relpath(file_path, path))

        logger.info(f"Created tar archive: {tar_path}")
    except FileNotFoundError as e:
        logger.error(f"File not found: {e}")
    except PermissionError as e:
        logger.error(f"Permission denied: {e}")
    except Exception as e:
        logger.error(f"An error occurred: {e}")


@main.group("package")
@click.pass_context
def package_cmd(_: click.Context) -> None:
    """:param ctx:"""
    pass


@package_cmd.command("create")
@click.option("-p", "--path", required=True, help="Path to package directory containing fedn.yaml")
@click.option("-n", "--name", required=False, default="package.tgz", help="Name of package tarball")
@click.pass_context
def create_cmd(_: click.Context, path: str, name: str) -> None:
    """Create compute package.

    Make a tar.gz archive of folder given by --path. The archive will be named --name.
    """
    try:
        path = os.path.abspath(path)
        yaml_file = os.path.join(path, "fedn.yaml")
        if not os.path.exists(yaml_file):
            logger.error(f"Could not find fedn.yaml in {path}")
            sys.exit(-1)

        create_tar_with_ignore(path, name)
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        sys.exit(-1)


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("--n_max", required=False, help="Number of items to list")
@package_cmd.command("list")
@click.pass_context
def list_packages(_: click.Context, protocol: str, host: str, port: str, token: str = None, n_max: int = None) -> None:
    """Return a list of packages.

    ------
    - count: number of packages
    - result: list of packages

    """
    headers = {}

    if n_max:
        headers["X-Limit"] = n_max

    response = get_response(protocol=protocol, host=host, port=port, endpoint="packages/", token=token, headers=headers, usr_api=False, usr_token=False)
    print_response(response, "packages", None)


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("-id", "--id", required=True, help="Package ID")
@package_cmd.command("get")
@click.pass_context
def get_package(_: click.Context, protocol: str, host: str, port: str, token: str = None, id: str = None) -> None:
    """Return a package with given id.

    ------
    - result: package with given id

    """
    response = get_response(protocol=protocol, host=host, port=port, endpoint=f"packages/{id}", token=token, headers={}, usr_api=False, usr_token=False)
    print_response(response, "package", id)



import json
import queue
import sys
import time
import traceback

from fedn.common.log_config import logger
from fedn.network.combiner.modelservice import ModelService, load_model_from_bytes


class ModelUpdateError(Exception):
    pass


class UpdateHandler:
    """Update handler.

    Responsible for receiving, loading and supplying client model updates.

    :param modelservice: A handle to the model service :class: `fedn.network.combiner.modelservice.ModelService`
    :type modelservice: class: `fedn.network.combiner.modelservice.ModelService`
    """

    def __init__(self, modelservice: ModelService) -> None:
        self.model_updates = queue.Queue()
        self.modelservice = modelservice

        self.model_id_to_model_data = {}

    def delete_model(self, model_update):
        self.modelservice.temp_model_storage.delete(model_update.model_update_id)
        logger.info("UPDATE HANDLER: Deleted model update {} from storage.".format(model_update.model_update_id))

    def next_model_update(self):
        """Get the next model update from the queue.

        :param helper: A helper object.
        :type helper: object
        :return: The model update.
        :rtype: fedn.network.grpc.fedn.proto.ModelUpdate
        """
        model_update = self.model_updates.get(block=False)
        return model_update

    def on_model_update(self, model_update):
        """Callback when a new client model update is recieved.

        Performs (optional) validation and pre-processing,
        and then puts the update id on the aggregation queue.
        Override in subclass as needed.

        :param model_update: fedn.network.grpc.fedn.proto.ModelUpdate
        :type model_id: str
        """
        try:
            logger.info("UPDATE HANDLER: callback received model update {}".format(model_update.model_update_id))

            # Validate the update and metadata
            valid_update = self._validate_model_update(model_update)
            if valid_update:
                # Push the model update to the processing queue
                self.model_updates.put(model_update)
            else:
                logger.warning("UPDATE HANDLER: Invalid model update, skipping.")
        except Exception as e:
            tb = traceback.format_exc()
            logger.error("UPDATE HANDLER: failed to receive model update: {}".format(e))
            logger.error(tb)
            pass

    def _validate_model_update(self, model_update):
        """Validate the model update.

        :param model_update: A ModelUpdate message.
        :type model_update: object
        :return: True if the model update is valid, False otherwise.
        :rtype: bool
        """
        try:
            data = json.loads(model_update.meta)["training_metadata"]
            _ = data["num_examples"]
        except KeyError:
            tb = traceback.format_exc()
            logger.error("UPDATE HANDLER: Invalid model update, missing metadata.")
            logger.error(tb)
            return False
        return True

    def load_model_update(self, model_update, helper):
        """Load the memory representation of the model update.

        Load the model update paramters and the
        associate metadata into memory.

        :param model_update: The model update.
        :type model_update: fedn.network.grpc.fedn.proto.ModelUpdate
        :param helper: A helper object.
        :type helper: fedn.utils.helpers.helperbase.Helper
        :return: A tuple of (parameters, metadata)
        :rtype: tuple
        """
        model_id = model_update.model_update_id
        model = self.load_model(helper, model_id)
        # Get relevant metadata
        metadata = json.loads(model_update.meta)
        if "config" in metadata.keys():
            # Used in Python client
            config = json.loads(metadata["config"])
        else:
            # Used in C++ client
            config = json.loads(model_update.config)
        training_metadata = metadata["training_metadata"]
        training_metadata["round_id"] = config["round_id"]

        return model, training_metadata

    def load_model_update_byte(self, model_update):
        """Load the memory representation of the model update.

        Load the model update paramters and the
        associate metadata into memory.

        :param model_update: The model update.
        :type model_update: fedn.network.grpc.fedn.proto.ModelUpdate
        :return: A tuple of parameters(bytes), metadata
        :rtype: tuple
        """
        model_id = model_update.model_update_id
        model = self.load_model_update_bytesIO(model_id).getbuffer()
        # Get relevant metadata
        metadata = json.loads(model_update.meta)
        if "config" in metadata.keys():
            # Used in Python client
            config = json.loads(metadata["config"])
        else:
            # Used in C++ client
            config = json.loads(model_update.config)
        training_metadata = metadata["training_metadata"]
        training_metadata["round_id"] = config["round_id"]

        return model, training_metadata

    def load_model(self, helper, model_id):
        """Load model update with id model_id into its memory representation.

        :param helper: An instance of :class: `fedn.utils.helpers.helpers.HelperBase`
        :type helper: class: `fedn.utils.helpers.helpers.HelperBase`
        :param model_id: The ID of the model update, UUID in str format
        :type model_id: str
        """
        model_bytesIO = self.load_model_update_bytesIO(model_id)
        if model_bytesIO:
            try:
                model = load_model_from_bytes(model_bytesIO.getbuffer(), helper)
            except IOError:
                logger.warning("UPDATE HANDLER: Failed to load model!")
        else:
            raise ModelUpdateError("Failed to load model.")

        return model

    def load_model_update_bytesIO(self, model_id, retry=3):
        """Load model update object and return it as BytesIO.

        :param model_id: The ID of the model
        :type model_id: str
        :param retry: number of times retrying load model update, defaults to 3
        :type retry: int, optional
        :return: Updated model
        :rtype: class: `io.BytesIO`
        """
        # Try reading model update from local disk/combiner memory
        model_str = self.modelservice.temp_model_storage.get(model_id)
        # And if we cannot access that, try downloading from the server
        if model_str is None:
            model_str = self.modelservice.get_model(model_id)
            # TODO: use retrying library
            tries = 0
            while tries < retry:
                tries += 1
                if not model_str or sys.getsizeof(model_str) == 80:
                    logger.warning("Model download failed. retrying")
                    time.sleep(1)
                    model_str = self.modelservice.get_model(model_id)

        return model_str

    def waitforit(self, config, buffer_size=100, polling_interval=0.1):
        """Defines the policy for how long the server should wait before starting to aggregate models.

        The policy is as follows:
            1. Wait a maximum of time_window time until the round times out.
            2. Terminate if a preset number of model updates (buffer_size) are in the queue.

        :param config: The round config object
        :type config: dict
        :param buffer_size: The number of model updates to wait for before starting aggregation, defaults to 100
        :type buffer_size: int, optional
        :param polling_interval: The polling interval, defaults to 0.1
        :type polling_interval: float, optional
        """
        time_window = float(config["round_timeout"])

        tt = 0.0
        while tt < time_window:
            if self.model_updates.qsize() >= buffer_size:
                break

            time.sleep(polling_interval)
            tt += polling_interval



import click

from fedn.network.combiner.hooks.hooks import serve

from .main import main


@main.group("hooks")
@click.pass_context
def hooks_cmd(ctx):
    """:param ctx:"""
    pass


@hooks_cmd.command("start")
@click.pass_context
def start_cmd(ctx):
    """:param ctx:"""
    click.echo("Started hooks container")
    serve()



import os

from flask import Flask, jsonify, request

from fedn.common.config import get_controller_config
from fedn.network.api import gunicorn_app
from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.shared import control
from fedn.network.api.v1 import _routes
from fedn.network.api.v1.graphql.schema import schema
from fedn.network.state import ReducerStateToString

custom_url_prefix = os.environ.get("FEDN_CUSTOM_URL_PREFIX", False)
app = Flask(__name__)
for bp in _routes:
    app.register_blueprint(bp)
    if custom_url_prefix:
        app.register_blueprint(bp, name=f"{bp.name}_custom", url_prefix=f"{custom_url_prefix}{bp.url_prefix}")


@app.route("/health", methods=["GET"])
def health_check():
    return "OK", 200


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/health", view_func=health_check, methods=["GET"])


@app.route("/api/v1/graphql", methods=["POST"])
def graphql_endpoint():
    data = request.get_json()

    if not data or "query" not in data:
        return jsonify({"error": "Missing query in request"}), 400

    # Execute the GraphQL query
    result = schema.execute(
        data["query"],
        variables=data.get("variables"),
        context_value={"request": request},  # Pass Flask request object as context if needed
    )

    # Format the result as a JSON response
    response = {"data": result.data}
    if result.errors:
        response["errors"] = [str(error) for error in result.errors]

    return jsonify(response)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/api/v1/graphql", view_func=graphql_endpoint, methods=["POST"])


@app.route("/get_controller_status", methods=["GET"])
@jwt_auth_required(role="admin")
def get_controller_status():
    """Get the status of the controller.
    return: The status as a json object.
    rtype: json
    """
    return jsonify({"state": ReducerStateToString(control.state())}), 200


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_controller_status", view_func=get_controller_status, methods=["GET"])


@app.route("/add_combiner", methods=["POST"])
@jwt_auth_required(role="combiner")
def add_combiner():
    """Add a combiner to the network.
    return: The response from the statestore.
    rtype: json
    """
    payload = {
        "success": False,
        "message": "Adding combiner via REST API is obsolete. Include statestore and object store config in combiner config.",
        "status": "abort",
    }

    return jsonify(payload), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/add_combiner", view_func=add_combiner, methods=["POST"])


# deprecated endpoints


@app.route("/get_model_trail", methods=["GET"])
@jwt_auth_required(role="admin")
def get_model_trail():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/models or the GraphQL API instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_model_trail", view_func=get_model_trail, methods=["GET"])


@app.route("/get_model_ancestors", methods=["GET"])
@jwt_auth_required(role="admin")
def get_model_ancestors():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/models/<id>/ancestors or the GraphQL API instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_model_ancestors", view_func=get_model_ancestors, methods=["GET"])


@app.route("/get_model_descendants", methods=["GET"])
@jwt_auth_required(role="admin")
def get_model_descendants():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/models/<id>/descendants or the GraphQL API instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_model_descendants", view_func=get_model_descendants, methods=["GET"])


@app.route("/list_models", methods=["GET"])
@jwt_auth_required(role="admin")
def list_models():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/models or the GraphQL API instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_models", view_func=list_models, methods=["GET"])


@app.route("/get_model", methods=["GET"])
@jwt_auth_required(role="admin")
def get_model():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/models/<id> or the GraphQL API instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_model", view_func=get_model, methods=["GET"])


@app.route("/list_clients", methods=["GET"])
@jwt_auth_required(role="admin")
def list_clients():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/clients instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_clients", view_func=list_clients, methods=["GET"])


@app.route("/get_active_clients", methods=["GET"])
@jwt_auth_required(role="admin")
def get_active_clients():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/clients instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_active_clients", view_func=get_active_clients, methods=["GET"])


@app.route("/list_combiners", methods=["GET"])
@jwt_auth_required(role="admin")
def list_combiners():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/combiners instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_combiners", view_func=list_combiners, methods=["GET"])


@app.route("/get_combiner", methods=["GET"])
@jwt_auth_required(role="admin")
def get_combiner():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/combiners/<id> instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_combiner", view_func=get_combiner, methods=["GET"])


@app.route("/list_rounds", methods=["GET"])
@jwt_auth_required(role="admin")
def list_rounds():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/rounds instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_rounds", view_func=list_rounds, methods=["GET"])


@app.route("/get_round", methods=["GET"])
@jwt_auth_required(role="admin")
def get_round():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/rounds/<id> instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_round", view_func=get_round, methods=["GET"])


@app.route("/start_session", methods=["GET", "POST"])
@jwt_auth_required(role="admin")
def start_session():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/sessions and /api/v1/sessions/start instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/start_session", view_func=start_session, methods=["GET", "POST"])


@app.route("/list_sessions", methods=["GET"])
@jwt_auth_required(role="admin")
def list_sessions():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/sessions or the GraphQL API instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_sessions", view_func=list_sessions, methods=["GET"])


@app.route("/get_session", methods=["GET"])
@jwt_auth_required(role="admin")
def get_session():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/sessions<id> or the GraphQL API instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_session", view_func=get_session, methods=["GET"])


@app.route("/set_active_package", methods=["PUT"])
@jwt_auth_required(role="admin")
def set_active_package():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/packages/active instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/set_active_package", view_func=set_active_package, methods=["PUT"])


@app.route("/set_package", methods=["POST"])
@jwt_auth_required(role="admin")
def set_package():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/packages instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/set_package", view_func=set_package, methods=["POST"])


@app.route("/get_package", methods=["GET"])
@jwt_auth_required(role="admin")
def get_package():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/packages/active or /api/v1/packages/<id> instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_package", view_func=get_package, methods=["GET"])


@app.route("/list_compute_packages", methods=["GET"])
@jwt_auth_required(role="admin")
def list_compute_packages():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/packages instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_compute_packages", view_func=list_compute_packages, methods=["GET"])


@app.route("/download_package", methods=["GET"])
@jwt_auth_required(role="client")
def download_package():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/packages/download instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/download_package", view_func=download_package, methods=["GET"])


@app.route("/get_package_checksum", methods=["GET"])
@jwt_auth_required(role="client")
def get_package_checksum():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/packages/checksum instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_package_checksum", view_func=get_package_checksum, methods=["GET"])


@app.route("/get_latest_model", methods=["GET"])
@jwt_auth_required(role="admin")
def get_latest_model():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/models/active instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_latest_model", view_func=get_latest_model, methods=["GET"])


@app.route("/set_current_model", methods=["PUT"])
@jwt_auth_required(role="admin")
def set_current_model():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/models/active instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/set_current_model", view_func=set_current_model, methods=["PUT"])


@app.route("/get_initial_model", methods=["GET"])
@jwt_auth_required(role="admin")
def get_initial_model():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/models/active instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_initial_model", view_func=get_initial_model, methods=["GET"])


@app.route("/set_initial_model", methods=["POST"])
@jwt_auth_required(role="admin")
def set_initial_model():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/models instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/set_initial_model", view_func=set_initial_model, methods=["POST"])


@app.route("/get_client_config", methods=["GET"])
@jwt_auth_required(role="admin")
def get_client_config():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/clients/config instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_client_config", view_func=get_client_config, methods=["GET"])


@app.route("/get_events", methods=["GET"])
@jwt_auth_required(role="admin")
def get_events():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/statuses instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_events", view_func=get_events, methods=["GET"])


@app.route("/list_validations", methods=["GET"])
@jwt_auth_required(role="admin")
def list_validations():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/validations or the GraphQL API instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_validations", view_func=list_validations, methods=["GET"])


@app.route("/add_client", methods=["POST"])
@jwt_auth_required(role="client")
def add_client():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/clients/add instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/add_client", view_func=add_client, methods=["POST"])


@app.route("/list_combiners_data", methods=["POST"])
@jwt_auth_required(role="admin")
def list_combiners_data():
    response = {
        "message": "This endpoint is deprecated. Use /api/v1/combiners/clients/count instead.",
    }
    return jsonify(response), 410


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_combiners_data", view_func=list_combiners_data, methods=["POST"])

# not implemented


@app.route("/delete_model_trail", methods=["GET", "POST"])
@jwt_auth_required(role="admin")
def delete_model_trail():
    return jsonify({"message": "Not implemented"}), 501


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/delete_model_trail", view_func=delete_model_trail, methods=["GET", "POST"])


def start_server_api():
    config = get_controller_config()
    port = config["port"]
    host = "0.0.0.0"
    debug = config["debug"]
    if debug:
        app.run(debug=debug, port=port, host=host)
    else:
        workers = os.cpu_count()
        gunicorn_app.run_gunicorn(app, host, port, workers)


if __name__ == "__main__":
    start_server_api()



"""Client module for handling client operations in the FEDn network."""

import io
import json
import os
import time
import uuid
from io import BytesIO
from typing import Dict, Optional, Tuple

from fedn.common.config import FEDN_CUSTOM_URL_PREFIX
from fedn.common.log_config import logger
from fedn.network.clients.fedn_client import ConnectToApiResult, FednClient, GrpcConnectionOptions
from fedn.network.combiner.modelservice import get_tmp_path
from fedn.utils.helpers.helpers import get_helper, save_metadata


def get_url(api_url: str, api_port: int) -> str:
    """Construct the URL for the API."""
    return f"{api_url}:{api_port}/{FEDN_CUSTOM_URL_PREFIX}" if api_port else f"{api_url}/{FEDN_CUSTOM_URL_PREFIX}"


class ClientOptions:
    """Options for configuring the client."""

    def __init__(self, name: str, package: str, preferred_combiner: Optional[str] = None, id: Optional[str] = None) -> None:
        """Initialize ClientOptions with validation."""
        self._validate(name, package)
        self.name = name
        self.package = package
        self.preferred_combiner = preferred_combiner
        self.client_id = id if id else str(uuid.uuid4())

    def _validate(self, name: str, package: str) -> None:
        """Validate the name and package."""
        if not isinstance(name, str) or len(name) == 0:
            raise ValueError("Name must be a string")
        if not isinstance(package, str) or len(package) == 0 or package not in ["local", "remote"]:
            raise ValueError("Package must be either 'local' or 'remote'")

    def to_json(self) -> Dict[str, Optional[str]]:
        """Convert ClientOptions to JSON."""
        return {
            "name": self.name,
            "client_id": self.client_id,
            "preferred_combiner": self.preferred_combiner,
            "package": self.package,
        }


class Client:
    """Client for interacting with the FEDn network."""

    def __init__(
        self,
        api_url: str,
        api_port: int,
        client_obj: ClientOptions,
        combiner_host: Optional[str] = None,
        combiner_port: Optional[int] = None,
        token: Optional[str] = None,
        package_checksum: Optional[str] = None,
        helper_type: Optional[str] = None,
    ) -> None:
        """Initialize the Client."""
        self.api_url = api_url
        self.api_port = api_port
        self.combiner_host = combiner_host
        self.combiner_port = combiner_port
        self.token = token
        self.client_obj = client_obj
        self.package_checksum = package_checksum
        self.helper_type = helper_type

        self.fedn_api_url = get_url(self.api_url, self.api_port)
        self.fedn_client: FednClient = FednClient()
        self.helper = None

    def _connect_to_api(self) -> Tuple[bool, Optional[dict]]:
        """Connect to the API and handle retries."""
        result = None
        response = None

        while not result or result == ConnectToApiResult.ComputePackageMissing:
            if result == ConnectToApiResult.ComputePackageMissing:
                logger.info("Retrying in 3 seconds")
                time.sleep(3)
            result, response = self.fedn_client.connect_to_api(self.fedn_api_url, self.token, self.client_obj.to_json())

        if result == ConnectToApiResult.Assigned:
            return True, response

        return False, None

    def start(self) -> None:
        """Start the client."""
        if self.combiner_host and self.combiner_port:
            combiner_config = GrpcConnectionOptions(host=self.combiner_host, port=self.combiner_port)
        else:
            result, combiner_config = self._connect_to_api()
            if not result:
                return

        if self.client_obj.package == "remote":
            result = self.fedn_client.init_remote_compute_package(url=self.fedn_api_url, token=self.token, package_checksum=self.package_checksum)
            if not result:
                return
        else:
            result = self.fedn_client.init_local_compute_package()
            if not result:
                return

        self.set_helper(combiner_config)

        result = self.fedn_client.init_grpchandler(config=combiner_config, client_name=self.client_obj.client_id, token=self.token)
        if not result:
            return

        logger.info("-----------------------------")

        self.fedn_client.set_train_callback(self.on_train)
        self.fedn_client.set_validate_callback(self.on_validation)
        self.fedn_client.set_predict_callback(self._process_prediction_request)

        self.fedn_client.set_name(self.client_obj.name)
        self.fedn_client.set_client_id(self.client_obj.client_id)

        self.fedn_client.run()

    def set_helper(self, response: Optional[GrpcConnectionOptions] = None) -> None:
        """Set the helper based on the response or default."""
        helper_type = response.helper_type if response else None
        helper_type_to_use = self.helper_type or helper_type or "numpyhelper"
        logger.info(f"Setting helper to: {helper_type_to_use}")
        self.helper = get_helper(helper_type_to_use)

    def on_train(self, in_model: BytesIO, client_settings: dict) -> Tuple[Optional[BytesIO], dict]:
        """Handle the training callback."""
        return self._process_training_request(in_model, client_settings)

    def on_validation(self, in_model: BytesIO) -> Optional[dict]:
        """Handle the validation callback."""
        return self._process_validation_request(in_model)

    def _process_training_request(self, in_model: BytesIO, client_settings: dict) -> Tuple[Optional[BytesIO], dict]:
        """Process a training (model update) request."""
        try:
            meta = {}
            inpath = self.helper.get_tmp_path()

            with open(inpath, "wb") as fh:
                fh.write(in_model.getbuffer())

            save_metadata(metadata=client_settings, filename=inpath)
            outpath = self.helper.get_tmp_path()
            tic = time.time()

            self.fedn_client.dispatcher.run_cmd(f"train {inpath} {outpath}")
            meta["exec_training"] = time.time() - tic

            with open(outpath, "rb") as fr:
                out_model = io.BytesIO(fr.read())

            with open(outpath + "-metadata", "r") as fh:
                training_metadata = json.loads(fh.read())

            logger.info(f"SETTING Training metadata: {training_metadata}")
            meta["training_metadata"] = training_metadata

            os.unlink(inpath)
            os.unlink(outpath)
            os.unlink(outpath + "-metadata")

        except Exception as e:
            logger.error(f"Could not process training request due to error: {e}")
            out_model = None
            meta = {"status": "failed", "error": str(e)}

        return out_model, meta

    def _process_validation_request(self, in_model: BytesIO) -> Optional[dict]:
        """Process a validation request."""
        try:
            inpath = self.helper.get_tmp_path()

            with open(inpath, "wb") as fh:
                fh.write(in_model.getbuffer())

            outpath = get_tmp_path()
            self.fedn_client.dispatcher.run_cmd(f"validate {inpath} {outpath}")

            with open(outpath, "r") as fh:
                metrics = json.loads(fh.read())

            os.unlink(inpath)
            os.unlink(outpath)

        except Exception as e:
            logger.warning(f"Validation failed with exception {e}")
            metrics = None

        return metrics

    def _process_prediction_request(self, in_model: BytesIO) -> Optional[dict]:
        """Process a prediction request."""
        try:
            inpath = self.helper.get_tmp_path()

            with open(inpath, "wb") as fh:
                fh.write(in_model.getbuffer())

            outpath = get_tmp_path()
            self.fedn_client.dispatcher.run_cmd(f"predict {inpath} {outpath}")

            with open(outpath, "r") as fh:
                metrics = json.loads(fh.read())

            os.unlink(inpath)
            os.unlink(outpath)

        except Exception as e:
            logger.warning(f"Prediction failed with exception {e}")
            metrics = None

        return metrics



import pytest
import pymongo

import itertools
import datetime
import uuid

from fedn.network.storage.dbconnection import DatabaseConnection

@pytest.fixture
def test_models():

    start_date = datetime.datetime(2021, 1, 4, 1, 2, 4)
    return [{"id":str(uuid.uuid4()), "key":"models", "model":"test_model1", "parent_model":"test_parent_model", "session_id":None, "committed_at":start_date - datetime.timedelta(days=1), "name":"test_name1", "active":True},
            {"id":str(uuid.uuid4()), "key":"models", "model":"test_model2", "parent_model":"test_parent_model", "session_id":None, "committed_at":start_date - datetime.timedelta(days=32), "name":"test_name2", "active":True},
            {"id":str(uuid.uuid4()), "key":"models", "model":"test_model3", "parent_model":"test_parent_model", "session_id":None, "committed_at":start_date - datetime.timedelta(days=23), "name":"test_name3", "active":True},
            {"id":str(uuid.uuid4()), "key":"models", "model":"test_model4", "parent_model":"test_parent_model", "session_id":None, "committed_at":start_date - datetime.timedelta(days=54), "name":"test_name4", "active":True},
            {"id":str(uuid.uuid4()), "key":"models", "model":"test_model5", "parent_model":"test_parent_model", "session_id":None, "committed_at":start_date - datetime.timedelta(days=25), "name":"test_name5", "active":True},
            {"id":str(uuid.uuid4()), "key":"models", "model":"test_model6", "parent_model":"test_parent_model2", "session_id":None, "committed_at":start_date - datetime.timedelta(days=16), "name":"test_name6", "active":True},
            {"id":str(uuid.uuid4()), "key":"models", "model":"test_model7", "parent_model":"test_parent_model2", "session_id":None, "committed_at":start_date - datetime.timedelta(days=27), "name":"test_name7", "active":True},
            {"id":str(uuid.uuid4()), "key":"models", "model":"test_model8", "parent_model":"test_parent_model2", "session_id":None, "committed_at":start_date - datetime.timedelta(days=48), "name":"test_name8", "active":True}]

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_models):
    for c in test_models:
        res, _ = mongo_connection.model_store.add(c)
        assert res == True

    for c in test_models:
        res, _ = postgres_connection.model_store.add(c)
        assert res == True

    for c in test_models:
        res, _ = sql_connection.model_store.add(c)
        assert res == True

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    # TODO:Clean up



@pytest.fixture
def options():
    sorting_keys = (#None, 
                        "name",
                        "committed_at",
                        #"invalid_key"
                        ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"parent_model":"test_parent_model2"}, {"active":False})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestModelStore:

    def test_add_update_delete(self, postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection):
        pass

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options: 
                res = db_1.model_store.list(*opt, **kwargs)
                count, gathered_models = res["count"], res["result"]

                res = db_2.model_store.list(*opt, **kwargs)
                count2, gathered_models2 = res["count"], res["result"]
                #TODO: The count is not equal to the number of clients in the list, but the number of clients returned by the query before skip and limit
                #It is not clear what is the intended behavior
                # assert(count == len(gathered_clients))
                # assert count == count2
                assert len(gathered_models) == len(gathered_models2)

                for i in range(len(gathered_models)):
                    #NOTE: id are not equal between the two databases, I think it is due to id being overwritten in the _id field
                    #assert gathered_models2[i]["id"] == gathered_models[i]["id"]
                    assert gathered_models2[i]["committed_at"] == gathered_models[i]["committed_at"]
                    assert gathered_models2[i]["model"] == gathered_models[i]["model"]
                    assert gathered_models2[i]["parent_model"] == gathered_models[i]["parent_model"]
                    assert gathered_models2[i]["session_id"] == gathered_models[i]["session_id"]
                    assert gathered_models2[i]["name"] == gathered_models[i]["name"]
                    assert gathered_models2[i]["active"] == gathered_models[i]["active"]
                    



from flask import Blueprint, jsonify, request

from fedn.common.log_config import logger
from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.shared import analytic_store
from fedn.network.api.v1.shared import api_version, get_typed_list_headers

bp = Blueprint("analytic", __name__, url_prefix=f"/api/{api_version}/analytics")


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_analytics():
    try:
        limit, skip, sort_key, sort_order = get_typed_list_headers(request.headers)
        kwargs = request.args.to_dict()

        response = analytic_store.list(limit, skip, sort_key, sort_order, **kwargs)

        return jsonify(response), 200
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/", methods=["POST"])
@jwt_auth_required(role="client")
def add_analytics():
    try:
        data = request.json if request.headers["Content-Type"] == "application/json" else request.form.to_dict()

        successful, result = analytic_store.add(data)
        response = result
        status_code: int = 201 if successful else 400

        return jsonify(response), status_code
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        return jsonify({"message": "An unexpected error occurred"}), 500



"""Base class for artifacts repository implementations."""

import abc
from typing import IO


class RepositoryBase(abc.ABC):
    """Base class for artifacts repository implementations."""

    @abc.abstractmethod
    def set_artifact(self, instance_name: str, instance: IO, bucket: str) -> None:
        """Set object with name instance_name.

        :param instance_name: The name of the object
        :type instance_name: str
        :param instance: The object
        :type instance: Any
        :param bucket: The bucket name
        :type bucket: str
        """
        raise NotImplementedError("Must be implemented by subclass")

    @abc.abstractmethod
    def get_artifact(self, instance_name: str, bucket: str) -> IO:
        """Retrieve object with name instance_name.

        :param instance_name: The name of the object to retrieve
        :type instance_name: str
        :param bucket: The bucket name
        :type bucket: str
        :return: The retrieved object
        :rtype: Any
        """
        raise NotImplementedError("Must be implemented by subclass")

    @abc.abstractmethod
    def get_artifact_stream(self, instance_name: str, bucket: str) -> IO:
        """Return a stream handler for object with name instance_name.

        :param instance_name: The name of the object
        :type instance_name: str
        :param bucket: The bucket name
        :type bucket: str
        :return: Stream handler for object instance_name
        :rtype: IO
        """
        raise NotImplementedError("Must be implemented by subclass")



import importlib
from abc import ABC, abstractmethod

from fedn.network.combiner.updatehandler import UpdateHandler

AGGREGATOR_PLUGIN_PATH = "fedn.network.combiner.aggregators.{}"


class AggregatorBase(ABC):
    """Abstract class defining an aggregator.

    :param control: A handle to the :class: `fedn.network.combiner.updatehandler.UpdateHandler`
    :type control: class: `fedn.network.combiner.updatehandler.UpdateHandler`
    """

    @abstractmethod
    def __init__(self, update_handler: UpdateHandler):
        """Initialize the aggregator."""
        self.name = self.__class__.__name__
        self.update_handler = update_handler

    @abstractmethod
    def combine_models(self, nr_expected_models=None, nr_required_models=1, helper=None, timeout=180, delete_models=True, parameters=None):
        """Routine for combining model updates. Implemented in subclass.

        :param nr_expected_models: Number of expected models. If None, wait for all models.
        :type nr_expected_models: int
        :param nr_required_models: Number of required models to combine.
        :type nr_required_models: int
        :param helper: A helper object.
        :type helper: :class: `fedn.utils.plugins.helperbase.HelperBase`
        :param timeout: Timeout in seconds to wait for models to be combined.
        :type timeout: int
        :param delete_models: Delete client models after combining.
        :type delete_models: bool
        :param parameters: Additional key-word arguments.
        :type parameters: dict
        :return: The global model and metadata
        :rtype: tuple
        """
        pass


def get_aggregator(aggregator_module_name, update_handler):
    """Return an instance of the helper class.

    :param helper_module_name: The name of the helper plugin module.
    :type helper_module_name: str
    :param storage: Model repository for :class: `fedn.network.combiner.Combiner`
    :type storage: class: `fedn.common.storage.s3.s3repo.S3ModelRepository`
    :param server: A handle to the Combiner class :class: `fedn.network.combiner.Combiner`
    :type server: class: `fedn.network.combiner.Combiner`
    :param modelservice: A handle to the model service :class: `fedn.network.combiner.modelservice.ModelService`
    :type modelservice: class: `fedn.network.combiner.modelservice.ModelService`
    :param control: A handle to the :class: `fedn.network.combiner.roundhandler.RoundHandler`
    :type control: class: `fedn.network.combiner.roundhandler.RoundHandler`
    :return: An aggregator instance.
    :rtype: class: `fedn.combiner.aggregators.AggregatorBase`
    """
    aggregator_plugin = AGGREGATOR_PLUGIN_PATH.format(aggregator_module_name)
    aggregator = importlib.import_module(aggregator_plugin)
    return aggregator.Aggregator(update_handler)



"""Interface for storing model objects and compute packages in S3 compatible storage."""

import datetime
import os
import uuid
from typing import Union

from fedn.common.log_config import logger
from fedn.network.storage.s3.miniorepository import MINIORepository
from fedn.network.storage.s3.saasrepository import SAASRepository


class Repository:
    """Interface for storing model objects and compute packages in S3 compatible storage."""

    def __init__(self, config: dict, init_buckets: bool = True, storage_type: str = "MINIO") -> None:
        """Initialize the repository.

        :param config: Configuration dictionary for credentials and bucket names.
        :type config: dict
        :param init_buckets: Whether to initialize buckets, defaults to True
        :type init_buckets: bool, optional
        :param storage_type: Type of storage to use, defaults to "MINIO"
        :type storage_type: str, optional
        """
        self.model_bucket = os.environ.get("FEDN_MODEL_BUCKET", config["storage_bucket"])
        self.context_bucket = os.environ.get("FEDN_CONTEXT_BUCKET", config["context_bucket"])
        try:
            self.prediction_bucket = os.environ.get("FEDN_PREDICTION_BUCKET", config["prediction_bucket"])
        except KeyError:
            self.prediction_bucket = "fedn-prediction"

        # TODO: Make a plug-in solution
        storage_type = os.environ.get("FEDN_STORAGE_TYPE", storage_type)
        if storage_type == "SAAS":
            self.client = SAASRepository(config)
        elif storage_type == "MINIO":
            self.client = MINIORepository(config)
        else:
            # Default to MinIO.
            self.client = MINIORepository(config)

        if init_buckets:
            self.client.create_bucket(self.context_bucket)
            self.client.create_bucket(self.model_bucket)
            self.client.create_bucket(self.prediction_bucket)

    def get_model(self, model_id: str) -> bytes:
        """Retrieve a model with id model_id.

        :param model_id: Unique identifier for model to retrieve.
        :type model_id: str
        :return: The model object
        :rtype: bytes
        """
        logger.info("Client {} trying to get model with id: {}".format(self.client.name, model_id))
        return self.client.get_artifact(model_id, self.model_bucket)

    def get_model_stream(self, model_id: str) -> bytes:
        """Retrieve a stream handle to model with id model_id.

        :param model_id: Unique identifier for model to retrieve.
        :type model_id: str
        :return: Handle to model object
        :rtype: bytes
        """
        logger.info("Client {} trying to get model with id: {}".format(self.client.name, model_id))
        return self.client.get_artifact_stream(model_id, self.model_bucket)

    def set_model(self, model: Union[bytes, str], is_file: bool = True) -> str:
        """Upload model object.

        :param model: The model object
        :type model: Union[bytes, str]
        :param is_file: True if model is a file name, else False
        :type is_file: bool, optional
        :return: id for the uploaded object
        :rtype: str
        """
        model_id = uuid.uuid4()

        try:
            self.client.set_artifact(str(model_id), model, bucket=self.model_bucket, is_file=is_file)
        except Exception as e:
            logger.error("Failed to upload model with ID {} to repository. Error: {}".format(model_id, e))
            raise Exception(f"Failed to upload model with ID {model_id} to repository.") from e
        return str(model_id)

    def delete_model(self, model_id: str) -> None:
        """Delete model.

        :param model_id: The id of the model to delete
        :type model_id: str
        """
        try:
            self.client.delete_artifact(model_id, bucket=self.model_bucket)
        except Exception as e:
            logger.error("Failed to delete model {} from repository. Error: {}".format(model_id, e))
            raise Exception(f"Failed to delete model {model_id} from repository.") from e

    def set_compute_package(self, name: str, compute_package: Union[bytes, str], is_file: bool = True) -> None:
        """Upload compute package.

        :param name: The name of the compute package.
        :type name: str
        :param compute_package: The compute package
        :type compute_package: Union[bytes, str]
        :param is_file: True if model is a file name, else False
        :type is_file: bool, optional
        """
        try:
            self.client.set_artifact(name, compute_package, bucket=self.context_bucket, is_file=is_file)
        except Exception as e:
            logger.error("Failed to write compute package to repository. Error: {}".format(e))
            raise Exception("Failed to write compute package to repository.") from e

    def get_compute_package(self, compute_package: str) -> bytes:
        """Retrieve compute package from object store.

        :param compute_package: The name of the compute package.
        :type compute_package: str
        :return: Compute package.
        :rtype: bytes
        """
        try:
            data = self.client.get_artifact(compute_package, bucket=self.context_bucket)
        except Exception as e:
            logger.error("Failed to get compute package from repository. Error: {}".format(e))
            raise Exception("Failed to get compute package from repository.") from e
        return data

    def delete_compute_package(self, compute_package: str) -> None:
        """Delete a compute package from storage.

        :param compute_package: The name of the compute package
        :type compute_package: str
        """
        try:
            self.client.delete_artifact(compute_package, bucket=self.context_bucket)
        except Exception as e:
            logger.error("Failed to delete compute package from repository. Error: {}".format(e))
            raise Exception("Failed to delete compute package from repository.") from e

    def presigned_put_url(self, bucket: str, object_name: str, expires: datetime.timedelta = datetime.timedelta(hours=1)) -> str:
        """Generate a presigned URL for an upload object request.

        :param bucket: The bucket name
        :type bucket: str
        :param object_name: The object name
        :type object_name: str
        :param expires: The time the URL is valid
        :type expires: datetime.timedelta
        :return: The URL
        :rtype: str
        """
        return self.client.client.presigned_put_object(bucket, object_name, expires)

    def presigned_get_url(self, bucket: str, object_name: str, expires: datetime.timedelta = datetime.timedelta(hours=1)) -> str:
        """Generate a presigned URL for a download object request.

        :param bucket: The bucket name
        :type bucket: str
        :param object_name: The object name
        :type object_name: str
        :param expires: The time the URL is valid
        :type expires: datetime.timedelta
        :return: The URL
        :rtype: str
        """
        return self.client.client.presigned_get_object(bucket, object_name, expires)



import click

from .main import main
from .shared import CONTROLLER_DEFAULTS, get_response, print_response


@main.group("round")
@click.pass_context
def round_cmd(ctx):
    """:param ctx:"""
    pass


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-s", "--session_id", required=False, help="Rounds in session with given session id")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("--n_max", required=False, help="Number of items to list")
@round_cmd.command("list")
@click.pass_context
def list_rounds(ctx, protocol: str, host: str, port: str, token: str = None, session_id: str = None, n_max: int = None):
    """Return:
    ------
    - count: number of rounds
    - result: list of rounds

    """
    headers = {}

    if n_max:
        headers["X-Limit"] = n_max

    if session_id:
        response = get_response(
            protocol=protocol,
            host=host,
            port=port,
            endpoint=f"rounds/?round_config.session_id={session_id}",
            token=token,
            headers=headers,
            usr_api=False,
            usr_token=False,
        )
    else:
        response = get_response(
            protocol=protocol,
            host=host,
            port=port,
            endpoint="rounds/",
            token=token,
            headers=headers,
            usr_api=False,
            usr_token=False,
        )
    print_response(response, "rounds", None)


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-id", "--id", required=True, help="Round ID")
@click.option("-t", "--token", required=False, help="Authentication token")
@round_cmd.command("get")
@click.pass_context
def get_round(ctx, protocol: str, host: str, port: str, token: str = None, id: str = None):
    """Return:
    ------
    - result: round with given id

    """
    response = get_response(protocol=protocol, host=host, port=port, endpoint=f"rounds/{id}", token=token, headers={}, usr_api=False, usr_token=False)
    print_response(response, "round", id)



import os

import click
import requests

from .main import main
from .shared import STUDIO_DEFAULTS, get_api_url, get_context, get_response, get_token, print_response, set_context

home_dir = os.path.expanduser("~")


@main.group("project")
@click.pass_context
def project_cmd(ctx):
    """:param ctx:"""
    pass


@click.option("-id", "--id", required=True, help="ID of project.")
@click.option("-p", "--protocol", required=False, default=STUDIO_DEFAULTS["protocol"], help="Communication protocol of studio (api)")
@click.option("-H", "--host", required=False, default=STUDIO_DEFAULTS["host"], help="Hostname of studio (api)")
@project_cmd.command("delete")
@click.pass_context
def delete_project(ctx, id: str = None, protocol: str = None, host: str = None):
    """Delete project with given ID."""
    # Check if project with given id exists
    studio_api = True

    response = get_response(protocol=protocol, host=host, port=None, endpoint=f"projects/{id}", token=None, headers={}, usr_api=studio_api, usr_token=False)
    if response.status_code == 200:
        if response.json().get("error"):
            click.secho(f"No project with id '{id}' exists.", fg="red")
        else:
            # Check if user wants to delete project with given id
            user_input = input(f"Are you sure you want to delete project with id {id} (y/n)?: ")
            if user_input == "y":
                url = get_api_url(protocol=protocol, host=host, port=None, endpoint=f"projects/delete/{id}", usr_api=studio_api)
                headers = {}

                _token = get_token(None, True)

                if _token:
                    headers["Authorization"] = _token
                # Call the authentication API
                try:
                    requests.delete(url, headers=headers)
                    click.secho(f"Project with slug {id} has been removed.", fg="green")
                except requests.exceptions.RequestException as e:
                    click.echo(str(e), fg="red")
                activate_project(None, protocol, host)
    else:
        click.secho(f"Unexpected error: {response.status_code}", fg="red")


@click.option("-n", "--name", required=False, default=None, help="Name of new projec.")
@click.option("-d", "--description", required=False, default=None, help="Description of new projec.")
@click.option("-p", "--protocol", required=False, default=STUDIO_DEFAULTS["protocol"], help="Communication protocol of studio (api)")
@click.option("-H", "--host", required=False, default=STUDIO_DEFAULTS["host"], help="Hostname of studio (api)")
@project_cmd.command("create")
@click.pass_context
def create_project(ctx, name: str = None, description: str = None, protocol: str = None, host: str = None):
    """Create project.
    :param ctx:
    """
    # Check if user can create project
    studio_api = True
    url = get_api_url(protocol=protocol, host=host, port=None, endpoint="projects/create", usr_api=studio_api)
    headers = {"Content-Type": "application/x-www-form-urlencoded"}

    _token = get_token(None, True)

    if _token:
        headers["Authorization"] = _token
    if name is None:
        name = input("Please enter a project name: ")
    if description is None:
        description = input("Please enter a project description (optional): ")
    if len(name) > 46 or len(description) >= 255:
        click.secho("Project name or description too long.", fg="red")
    else:
        # Call the authentication API
        try:
            requests.post(url, data={"name": name, "description": description}, headers=headers)
        except requests.exceptions.RequestException as e:
            click.secho(str(e), fg="red")
        click.secho("Project created.", fg="green")


@click.option("-p", "--protocol", required=False, default=STUDIO_DEFAULTS["protocol"], help="Communication protocol of studio (api)")
@click.option("-H", "--host", required=False, default=STUDIO_DEFAULTS["host"], help="Hostname of studio (api)")
@project_cmd.command("list")
@click.pass_context
def list_projects(ctx, protocol: str = None, host: str = None):
    """Return:
    ------
    - result: list of projects

    """
    studio_api = True

    response = get_response(protocol=protocol, host=host, port=None, endpoint="projects", token=None, headers={}, usr_api=studio_api, usr_token=True)

    if response.status_code == 200:
        response_json = response.json()
        if len(response_json) > 0:
            context_path = os.path.join(home_dir, ".fedn")
            context_data = get_context(context_path)
            active_project = context_data.get("Active project id")

            for i in response_json:
                project_name = i.get("slug")
                if project_name == active_project:
                    click.secho(f"{project_name} (active)", fg="green")
                else:
                    click.secho(project_name)
    else:
        click.secho(f"Unexpected error: {response.status_code}", fg="red")


@click.option("-id", "--id", required=True, help="ID of project.")
@click.option("-p", "--protocol", required=False, default=STUDIO_DEFAULTS["protocol"], help="Communication protocol of studio (api)")
@click.option("-H", "--host", required=False, default=STUDIO_DEFAULTS["host"], help="Hostname of studio (api)")
@project_cmd.command("get")
@click.pass_context
def get_project(ctx, id: str = None, protocol: str = None, host: str = None):
    """Return:
    ------
    - result: project with given id

    """
    studio_api = True

    response = get_response(protocol=protocol, host=host, port=None, endpoint=f"projects/{id}", token=None, headers={}, usr_api=studio_api, usr_token=False)

    if response.status_code == 200:
        response_json = response.json()

        if response_json.get("error"):
            click.secho(f"No project with id '{id}' exists.", fg="red")
        else:
            print_response(response, "project", True)
    else:
        click.secho(f"Unexpected error: {response.status_code}", fg="red")


@click.option("-id", "--id", required=True, help="id name of project.")
@click.option("-p", "--protocol", required=False, default=STUDIO_DEFAULTS["protocol"], help="Communication protocol of studio (api)")
@click.option("-H", "--host", required=False, default=STUDIO_DEFAULTS["host"], help="Hostname of studio (api)")
@project_cmd.command("set-context")
@click.pass_context
def set_active_project(ctx, id: str = None, protocol: str = None, host: str = None):
    """Set active project.

    :param ctx:
    :param id:
    """
    activate_project(id, protocol, host)


def activate_project(id: str = None, protocol: str = None, host: str = None):
    """Sets project with give ID as active by updating context file."""
    studio_api = True
    headers_projects = {}
    context_path = os.path.join(home_dir, ".fedn")
    context_data = get_context(context_path)

    user_access_token = context_data.get("User tokens").get("access")

    response_projects = get_response(
        protocol=protocol, host=host, port=None, endpoint="projects", token=user_access_token, headers=headers_projects, usr_api=studio_api, usr_token=False
    )
    if response_projects.status_code == 200:
        projects_response_json = response_projects.json()
        if len(projects_response_json) > 0:
            if id is None:
                headers_projects["X-Project-Slug"] = projects_response_json[0].get("slug")
                id = projects_response_json[0].get("slug")
            else:
                for i in projects_response_json:
                    project_found = False
                    if i.get("slug") == id:
                        project_found = True
                        headers_projects["X-Project-Slug"] = i.get("slug")
                if not project_found:
                    click.secho(f"No project found with id {id}", fg="red")
                    return
            controller_url = f"{protocol}://{host}/{id}-fedn-reducer"

            response_project_tokens = get_response(
                protocol=protocol,
                host=host,
                port=None,
                endpoint="admin-token",
                token=user_access_token,
                headers=headers_projects,
                usr_api=studio_api,
                usr_token=False,
            )
            if response_project_tokens.status_code == 200:
                project_tokens = response_project_tokens.json()
                context_data["Active project tokens"] = project_tokens
                context_data["Active project id"] = id
                context_data["Active project url"] = controller_url

                set_context(context_path, context_data)

                click.secho(f"Project with slug {id} is now active.", fg="green")
            else:
                click.secho(f"Unexpected error: {response_project_tokens.status_code}", fg="red")
        else:
            click.echo("No projects available to set current context.")
    else:
        click.secho(f"Unexpected error: {response_projects.status_code}", fg="red")


def no_project_exists(response) -> bool:
    """Returns true if no project exists."""
    response_json = response.json()
    print(response_json)
    if type(response_json) is list:
        return False
    elif response_json.get("error"):
        return True
    return False



import json
import os

import grpc

import fedn.network.grpc.fedn_pb2 as fedn
import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.log_config import logger
from fedn.network.combiner.modelservice import bytesIO_request_generator, model_as_bytesIO, unpack_model
from fedn.network.combiner.updatehandler import UpdateHandler

CHUNK_SIZE = 1024 * 1024


class CombinerHookInterface:
    """Combiner to server function hooks client."""

    def __init__(self):
        """Initialize CombinerHookInterface client."""
        self.hook_service_host = os.getenv("HOOK_SERVICE_HOST", "hook:12081")
        self.channel = grpc.insecure_channel(
            self.hook_service_host,
            options=[
                ("grpc.keepalive_time_ms", 30000),  # 30 seconds ping interval
                ("grpc.keepalive_timeout_ms", 5000),  # 5 seconds timeout for a response
                ("grpc.keepalive_permit_without_calls", 1),  # allow keepalives even with no active calls
                ("grpc.enable_retries", 1),  # automatic retries
                ("grpc.initial_reconnect_backoff_ms", 1000),  # initial delay before retrying
                ("grpc.max_reconnect_backoff_ms", 5000),  # maximum delay before retrying
            ],
        )
        self.stub = rpc.FunctionServiceStub(self.channel)

    def provided_functions(self, server_functions: str):
        """Communicates to hook container and asks which functions are available.

        :param server_functions: String version of an implementation of the ServerFunctionsBase interface.
        :type server_functions: :str:
        :return: dictionary specifing which functions are implemented.
        :rtype: dict
        """
        try:
            request = fedn.ProvidedFunctionsRequest(function_code=server_functions)

            response = self.stub.HandleProvidedFunctions(request)
            return response.available_functions
        except grpc.RpcError as rpc_error:
            if rpc_error.code() == grpc.StatusCode.UNAVAILABLE:
                logger.warning(f"Server-functions container is unavailable; using default implementations: {rpc_error}")
            else:
                logger.error(f"gRPC error: {rpc_error.code().name} - {rpc_error.details()}")
            return {}
        except Exception as e:
            logger.error(f"Unexpected error communicating with hooks container: {e}")
            return {}

    def client_settings(self, global_model) -> dict:
        """Communicates to hook container to get a client config.

        :param global_model: The global model that will be distributed to clients.
        :type global_model: :bytes:
        :return: config that will be distributed to clients.
        :rtype: dict
        """
        request_function = fedn.ClientConfigRequest
        args = {}
        model = model_as_bytesIO(global_model)
        response = self.stub.HandleClientConfig(bytesIO_request_generator(mdl=model, request_function=request_function, args=args))
        return json.loads(response.client_settings)

    def client_selection(self, clients: list) -> list:
        request = fedn.ClientSelectionRequest(client_ids=json.dumps(clients))
        response = self.stub.HandleClientSelection(request)
        return json.loads(response.client_ids)

    def aggregate(self, previous_global, update_handler: UpdateHandler, helper, delete_models: bool):
        """Aggregation call to the hook functions. Sends models in chunks, then asks for aggregation.

        :param global_model: The global model that will be distributed to clients.
        :type global_model: :bytes:
        :return: config that will be distributed to clients.
        :rtype: dict
        """
        data = {}
        data["time_model_load"] = 0.0
        data["time_model_aggregation"] = 0.0
        # send previous global
        request_function = fedn.StoreModelRequest
        args = {"id": "global_model"}
        response = self.stub.HandleStoreModel(bytesIO_request_generator(mdl=previous_global, request_function=request_function, args=args))
        logger.info(f"Store model response: {response.status}")
        # send client models and metadata
        nr_updates = 0
        while not update_handler.model_updates.empty():
            logger.info("Getting next model update from queue.")
            update = update_handler.next_model_update()
            metadata = json.loads(update.meta)["training_metadata"]
            model = update_handler.load_model_update_bytesIO(update.model_update_id)
            # send metadata
            client_id = update.sender.client_id
            request = fedn.ClientMetaRequest(metadata=json.dumps(metadata), client_id=client_id)
            response = self.stub.HandleMetadata(request)
            # send client model
            args = {"id": client_id}
            request_function = fedn.StoreModelRequest
            response = self.stub.HandleStoreModel(bytesIO_request_generator(mdl=model, request_function=request_function, args=args))
            logger.info(f"Store model response: {response.status}")
            nr_updates += 1
            if delete_models:
                # delete model from disk
                update_handler.delete_model(model_update=update)
        # ask for aggregation
        request = fedn.AggregationRequest(aggregate="aggregate")
        response_generator = self.stub.HandleAggregation(request)
        data["nr_aggregated_models"] = nr_updates
        model, _ = unpack_model(response_generator, helper)
        return model, data



"""GrpcHandler class for handling GRPC connections and operations."""

import json
import os
import time
from datetime import datetime, timezone
from io import BytesIO
from typing import Any, Callable, Optional, Union

import grpc
from google.protobuf.json_format import MessageToJson

import fedn.network.grpc.fedn_pb2 as fedn
import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.config import FEDN_AUTH_SCHEME
from fedn.common.log_config import logger
from fedn.network.combiner.modelservice import upload_request_generator

# Keepalive settings: these help keep the connection open for long-lived clients
KEEPALIVE_TIME_MS = 1 * 1000  # send keepalive ping every 60 seconds
KEEPALIVE_TIMEOUT_MS = 30 * 1000  # wait 20 seconds for keepalive ping ack before considering connection dead
KEEPALIVE_PERMIT_WITHOUT_CALLS = True  # allow keepalive pings even when there are no RPCs
MAX_CONNECTION_IDLE_MS = 30000
MAX_CONNECTION_AGE_GRACE_MS = "INT_MAX"  # keep connection open indefinitely
CLIENT_IDLE_TIMEOUT_MS = 30000

GRPC_OPTIONS = [
    ("grpc.keepalive_time_ms", KEEPALIVE_TIME_MS),
    ("grpc.keepalive_timeout_ms", KEEPALIVE_TIMEOUT_MS),
    ("grpc.keepalive_permit_without_calls", KEEPALIVE_PERMIT_WITHOUT_CALLS),
    ("grpc.http2.max_pings_without_data", 0),  # unlimited pings without data
    ("grpc.max_connection_idle_ms", MAX_CONNECTION_IDLE_MS),
    ("grpc.max_connection_age_grace_ms", MAX_CONNECTION_AGE_GRACE_MS),
    ("grpc.client_idle_timeout_ms", CLIENT_IDLE_TIMEOUT_MS),
]

GRPC_SECURE_PORT = 443


class GrpcAuth(grpc.AuthMetadataPlugin):
    """GRPC authentication plugin."""

    def __init__(self, key: str) -> None:
        """Initialize GrpcAuth with a key."""
        self._key = key

    def __call__(self, context: grpc.AuthMetadataContext, callback: grpc.AuthMetadataPluginCallback) -> None:
        """Add authorization metadata to the GRPC call."""
        callback((("authorization", f"{FEDN_AUTH_SCHEME} {self._key}"),), None)


class GrpcHandler:
    """Handler for GRPC connections and operations."""

    def __init__(self, host: str, port: int, name: str, token: str, combiner_name: str) -> None:
        """Initialize the GrpcHandler."""
        self.metadata = [
            ("client", name),
            ("grpc-server", combiner_name),
        ]
        self.host = host
        self.port = port
        self.token = token

        self._init_channel(host, port, token)
        self._init_stubs()

    def _init_stubs(self) -> None:
        """Initialize GRPC stubs."""
        self.connectorStub = rpc.ConnectorStub(self.channel)
        self.combinerStub = rpc.CombinerStub(self.channel)
        self.modelStub = rpc.ModelServiceStub(self.channel)

    def _init_channel(self, host: str, port: int, token: str) -> None:
        """Initialize the GRPC channel."""
        if port == GRPC_SECURE_PORT:
            self._init_secure_channel(host, port, token)
        else:
            self._init_insecure_channel(host, port)

    def _init_secure_channel(self, host: str, port: int, token: str) -> None:
        """Initialize a secure GRPC channel."""
        url = f"{host}:{port}"
        logger.info(f"Connecting (GRPC) to {url}")

        if os.getenv("FEDN_GRPC_ROOT_CERT_PATH"):
            logger.info("Using root certificate from environment variable for GRPC channel.")
            with open(os.environ["FEDN_GRPC_ROOT_CERT_PATH"], "rb") as f:
                credentials = grpc.ssl_channel_credentials(f.read())
            self.channel = grpc.secure_channel(f"{host}:{port}", credentials)
            return

        credentials = grpc.ssl_channel_credentials()
        auth_creds = grpc.metadata_call_credentials(GrpcAuth(token))
        self.channel = grpc.secure_channel(
            f"{host}:{port}",
            grpc.composite_channel_credentials(credentials, auth_creds),
            options=GRPC_OPTIONS,
        )

    def _init_insecure_channel(self, host: str, port: int) -> None:
        """Initialize an insecure GRPC channel."""
        url = f"{host}:{port}"
        logger.info(f"Connecting (GRPC) to {url}")
        self.channel = grpc.insecure_channel(
            url,
            options=GRPC_OPTIONS,
        )

    def heartbeat(self, client_name: str, client_id: str) -> fedn.Response:
        """Send a heartbeat to the combiner.

        :return: Response from the combiner.
        :rtype: fedn.Response
        """
        heartbeat = fedn.Heartbeat(sender=fedn.Client(name=client_name, role=fedn.CLIENT, client_id=client_id))

        try:
            logger.info("Sending heartbeat to combiner")
            response = self.connectorStub.SendHeartbeat(heartbeat, metadata=self.metadata)
        except grpc.RpcError as e:
            logger.error(f"GRPC (SendHeartbeat): An error occurred: {e}")
            raise e
        except Exception as e:
            logger.error(f"GRPC (SendHeartbeat): An error occurred: {e}")
            raise e
        return response

    def send_heartbeats(self, client_name: str, client_id: str, update_frequency: float = 2.0) -> None:
        """Send heartbeats to the combiner at regular intervals."""
        send_heartbeat = True
        while send_heartbeat:
            try:
                response = self.heartbeat(client_name, client_id)
            except grpc.RpcError as e:
                self._handle_grpc_error(e, "SendHeartbeat", lambda: self.send_heartbeats(client_name, client_id, update_frequency))
                return
            except Exception as e:
                self._handle_unknown_error(e, "SendHeartbeat", lambda: self.send_heartbeats(client_name, client_id, update_frequency))
                return
            if isinstance(response, fedn.Response):
                logger.info("Heartbeat successful.")
            else:
                logger.error("Heartbeat failed.")
                send_heartbeat = False
            time.sleep(update_frequency)

    def listen_to_task_stream(self, client_name: str, client_id: str, callback: Callable[[Any], None]) -> None:
        """Subscribe to the model update request stream."""
        r = fedn.ClientAvailableMessage()
        r.sender.name = client_name
        r.sender.role = fedn.CLIENT
        r.sender.client_id = client_id

        try:
            logger.info("Listening to task stream.")
            for request in self.combinerStub.TaskStream(r, metadata=self.metadata):
                if request.sender.role == fedn.COMBINER:
                    self.send_status(
                        "Received request from combiner.",
                        log_level=fedn.LogLevel.AUDIT,
                        type=request.type,
                        request=request,
                        sesssion_id=request.session_id,
                        sender_name=client_name,
                    )

                    logger.info(f"Received task request of type {request.type} for model_id {request.model_id}")

                    callback(request)

        except grpc.RpcError as e:
            logger.error(f"GRPC (TaskStream): An error occurred: {e}")
            self._handle_grpc_error(e, "TaskStream", lambda: self.listen_to_task_stream(client_name, client_id, callback))
        except Exception as e:
            logger.error(f"GRPC (TaskStream): An error occurred: {e}")
            self._handle_unknown_error(e, "TaskStream", lambda: self.listen_to_task_stream(client_name, client_id, callback))

    def send_status(
        self,
        msg: str,
        log_level: fedn.LogLevel = fedn.LogLevel.INFO,
        type: Optional[str] = None,
        request: Optional[Union[fedn.ModelUpdate, fedn.ModelValidation, fedn.TaskRequest]] = None,
        sesssion_id: Optional[str] = None,
        sender_name: Optional[str] = None,
    ) -> None:
        """Send status message.

        :param msg: The message to send.
        :type msg: str
        :param log_level: The log level of the message.
        :type log_level: fedn.LogLevel.INFO, fedn.LogLevel.WARNING, fedn.LogLevel.ERROR
        :param type: The type of the message.
        :type type: str
        :param request: The request message.
        :type request: fedn.Request
        """
        status = fedn.Status()
        status.timestamp.GetCurrentTime()
        status.sender.name = sender_name
        status.sender.role = fedn.CLIENT
        status.log_level = log_level
        status.status = str(msg)
        status.session_id = sesssion_id

        if type is not None:
            status.type = type

        if request is not None:
            status.data = MessageToJson(request)

        try:
            logger.info("Sending status message to combiner.")
            _ = self.connectorStub.SendStatus(status, metadata=self.metadata)
        except grpc.RpcError as e:
            self._handle_grpc_error(e, "SendStatus", lambda: self.send_status(msg, log_level, type, request, sesssion_id, sender_name))
        except Exception as e:
            logger.error(f"GRPC (SendStatus): An error occurred: {e}")
            self._handle_unknown_error(e, "SendStatus", lambda: self.send_status(msg, log_level, type, request, sesssion_id, sender_name))

    def get_model_from_combiner(self, id: str, client_id: str, timeout: int = 20) -> Optional[BytesIO]:
        """Fetch a model from the assigned combiner.

        Downloads the model update object via a gRPC streaming channel.

        :param id: The id of the model update object.
        :type id: str
        :param client_id: The id of the client.
        :type client_id: str
        :param timeout: The timeout for the request.
        :type timeout: int
        :return: The model update object.
        :rtype: Optional[BytesIO]
        """
        data = BytesIO()
        time_start = time.time()
        request = fedn.ModelRequest(id=id)
        request.sender.client_id = client_id
        request.sender.role = fedn.CLIENT

        try:
            logger.info("Downloading model from combiner.")
            for part in self.modelStub.Download(request, metadata=self.metadata):
                if part.status == fedn.ModelStatus.IN_PROGRESS:
                    data.write(part.data)

                if part.status == fedn.ModelStatus.OK:
                    return data

                if part.status == fedn.ModelStatus.FAILED:
                    return None

                if part.status == fedn.ModelStatus.UNKNOWN:
                    if time.time() - time_start >= timeout:
                        return None
                    continue
        except grpc.RpcError as e:
            return self._handle_grpc_error(e, "Download", lambda: self.get_model_from_combiner(id, client_id, timeout))
        except Exception as e:
            logger.error(f"GRPC (Download): An error occurred: {e}")
            self._handle_unknown_error(e, "Download", lambda: self.get_model_from_combiner(id, client_id, timeout))
        return data

    def send_model_to_combiner(self, model: BytesIO, id: str) -> Optional[BytesIO]:
        """Send a model update to the assigned combiner.

        Uploads the model updated object via a gRPC streaming channel, Upload.

        :param model: The model update object.
        :type model: BytesIO
        :param id: The id of the model update object.
        :type id: str
        :return: The model update object.
        :rtype: Optional[BytesIO]
        """
        if not isinstance(model, BytesIO):
            bt = BytesIO()

            for d in model.stream(32 * 1024):
                bt.write(d)
        else:
            bt = model

        bt.seek(0, 0)

        try:
            logger.info("Uploading model to combiner.")
            result = self.modelStub.Upload(upload_request_generator(bt, id), metadata=self.metadata)
        except grpc.RpcError as e:
            return self._handle_grpc_error(e, "Upload", lambda: self.send_model_to_combiner(model, id))
        except Exception as e:
            logger.error(f"GRPC (Upload): An error occurred: {e}")
            self._handle_unknown_error(e, "Upload", lambda: self.send_model_to_combiner(model, id))
        return result

    def create_update_message(
        self,
        sender_name: str,
        model_id: str,
        model_update_id: str,
        receiver_name: str,
        receiver_role: fedn.Role,
        meta: dict,
    ) -> fedn.ModelUpdate:
        """Create an update message."""
        update = fedn.ModelUpdate()
        update.sender.name = sender_name
        update.sender.role = fedn.CLIENT
        update.sender.client_id = self.metadata[0][1]
        update.receiver.name = receiver_name
        update.receiver.role = receiver_role
        update.model_id = model_id
        update.model_update_id = model_update_id
        update.timestamp = str(datetime.now(timezone.utc))
        update.meta = json.dumps(meta)

        return update

    def create_validation_message(
        self,
        sender_name: str,
        receiver_name: str,
        receiver_role: fedn.Role,
        model_id: str,
        metrics: str,
        correlation_id: str,
        session_id: str,
    ) -> fedn.ModelValidation:
        """Create a validation message."""
        validation = fedn.ModelValidation()
        validation.sender.name = sender_name
        validation.sender.role = fedn.CLIENT
        validation.receiver.name = receiver_name
        validation.receiver.role = receiver_role
        validation.model_id = model_id
        validation.data = metrics
        validation.timestamp.GetCurrentTime()
        validation.correlation_id = correlation_id
        validation.session_id = session_id

        return validation

    def create_prediction_message(
        self,
        sender_name: str,
        receiver_name: str,
        receiver_role: fedn.Role,
        model_id: str,
        prediction_output: str,
        correlation_id: str,
        session_id: str,
    ) -> fedn.ModelPrediction:
        """Create a prediction message."""
        prediction = fedn.ModelPrediction()
        prediction.sender.name = sender_name
        prediction.sender.role = fedn.CLIENT
        prediction.receiver.name = receiver_name
        prediction.receiver.role = receiver_role
        prediction.model_id = model_id
        prediction.data = prediction_output
        prediction.timestamp.GetCurrentTime()
        prediction.correlation_id = correlation_id
        prediction.prediction_id = session_id

        return prediction

    def send_model_update(self, update: fedn.ModelUpdate) -> bool:
        """Send a model update to the combiner."""
        try:
            logger.info("Sending model update to combiner.")
            _ = self.combinerStub.SendModelUpdate(update, metadata=self.metadata)
        except grpc.RpcError as e:
            return self._handle_grpc_error(e, "SendModelUpdate", lambda: self.send_model_update(update))
        except Exception as e:
            logger.error(f"GRPC (SendModelUpdate): An error occurred: {e}")
            self._handle_unknown_error(e, "SendModelUpdate", lambda: self.send_model_update(update))
        return True

    def send_model_validation(self, validation: fedn.ModelValidation) -> bool:
        """Send a model validation to the combiner."""
        try:
            logger.info("Sending model validation to combiner.")
            _ = self.combinerStub.SendModelValidation(validation, metadata=self.metadata)
        except grpc.RpcError as e:
            return self._handle_grpc_error(
                e,
                "SendModelValidation",
                lambda: self.send_model_validation(validation),
            )
        except Exception as e:
            logger.error(f"GRPC (SendModelValidation): An error occurred: {e}")
            self._handle_unknown_error(e, "SendModelValidation", lambda: self.send_model_validation(validation))
        return True

    def send_model_prediction(self, prediction: fedn.ModelPrediction) -> bool:
        """Send a model prediction to the combiner."""
        try:
            logger.info("Sending model prediction to combiner.")
            _ = self.combinerStub.SendModelPrediction(prediction, metadata=self.metadata)
        except grpc.RpcError as e:
            return self._handle_grpc_error(
                e,
                "SendModelPrediction",
                lambda: self.send_model_prediction(prediction),
            )
        except Exception as e:
            logger.error(f"GRPC (SendModelPrediction): An error occurred: {e}")
            self._handle_unknown_error(e, "SendModelPrediction", lambda: self.send_model_prediction(prediction))
        return True

    def _handle_grpc_error(self, e: grpc.RpcError, method_name: str, sender_function: Callable) -> Optional[Callable]:
        """Handle GRPC errors."""
        status_code = e.code()
        if status_code == grpc.StatusCode.UNAVAILABLE:
            logger.warning(f"GRPC ({method_name}): server unavailable. Retrying in 5 seconds.")
            time.sleep(5)
            return sender_function()
        if status_code == grpc.StatusCode.CANCELLED:
            logger.warning(f"GRPC ({method_name}): connection cancelled. Retrying in 5 seconds.")
            time.sleep(5)
            return sender_function()
        if status_code == grpc.StatusCode.UNAUTHENTICATED:
            details = e.details()
            if details == "Token expired":
                logger.warning(f"GRPC ({method_name}): Token expired.")
                raise e
        if status_code == grpc.StatusCode.UNKNOWN:
            logger.warning(f"GRPC ({method_name}): An unknown error occurred: {e}.")
            details = e.details()
            if details == "Stream removed":
                logger.warning(f"GRPC ({method_name}): Stream removed. Reconnecting")
                self._disconnect()
                self._init_channel(self.host, self.port, self.token)
                self._init_stubs()
                return sender_function()
            raise e
        self._disconnect()
        logger.error(f"GRPC ({method_name}): An error occurred: {e}")
        raise e

    def _handle_unknown_error(self, e: Exception, method_name: str, sender_function: Callable) -> Optional[Callable]:
        """Handle unknown errors."""
        logger.warning(f"GRPC ({method_name}): An unknown error occurred: {e}.")
        if isinstance(e, ValueError):
            # ValueError is raised when the channel is closed
            self._disconnect()
            logger.warning(f"GRPC ({method_name}): Reconnecting to channel.")
            # recreate the channel
            self._init_channel(self.host, self.port, self.token)
            self._init_stubs()
            return sender_function()
        raise e

    def _disconnect(self) -> None:
        """Disconnect from the combiner."""
        self.channel.close()
        logger.info("GRPC channel closed.")



import os

import click
import requests
import yaml

from fedn.common.log_config import logger

CONTROLLER_DEFAULTS = {"protocol": "http", "host": "localhost", "port": 8092, "debug": False}

STUDIO_DEFAULTS = {"protocol": "https", "host": "fedn.scaleoutsystems.com"}

COMBINER_DEFAULTS = {"discover_host": "localhost", "discover_port": 8092, "host": "localhost", "port": 12080, "name": "combiner", "max_clients": 30}

CLIENT_DEFAULTS = {
    "discover_host": "localhost",
    "discover_port": 8092,
}

API_VERSION = "v1"

home_dir = os.path.expanduser("~")


def apply_config(path: str, config: dict):
    """Parse client config from file.

    Override configs from the CLI with settings in config file.

    :param config: Client config (dict).
    """
    with open(path, "r") as file:
        try:
            settings = dict(yaml.safe_load(file))
        except Exception:
            logger.error("Failed to read config from settings file, exiting.")
            return

    for key, val in settings.items():
        config[key] = val


def get_api_url(protocol: str, host: str, port: str, endpoint: str, usr_api: bool) -> str:
    if usr_api:
        _url = os.environ.get("FEDN_STUDIO_URL")
        _protocol = protocol or os.environ.get("FEDN_STUDIO_PROTOCOL") or STUDIO_DEFAULTS["protocol"]
        _host = host or os.environ.get("FEDN_STUDIO_HOST") or STUDIO_DEFAULTS["host"]

        if _url is None:
            return f"{_protocol}://{_host}/api/{API_VERSION}/{endpoint}"

        return f"{_url}/api/{API_VERSION}/{endpoint}"
    else:
        _url = os.environ.get("FEDN_CONTROLLER_URL")
        _protocol = protocol or os.environ.get("FEDN_CONTROLLER_PROTOCOL") or CONTROLLER_DEFAULTS["protocol"]
        _host = host or os.environ.get("FEDN_CONTROLLER_HOST") or CONTROLLER_DEFAULTS["host"]
        _port = port or os.environ.get("FEDN_CONTROLLER_PORT") or CONTROLLER_DEFAULTS["port"]

        if _url is None:
            context_path = os.path.join(home_dir, ".fedn")
            try:
                context_data = get_context(context_path)
                _url = context_data.get("Active project url")
            except Exception as e:
                click.echo(f"Encountered error {e}. Make sure you are logged in and have activated a project. Using controller defaults instead.", fg="red")
                _url = f"{_protocol}://{_host}:{_port}"

    return f"{_url}/api/{API_VERSION}/{endpoint}"


def get_token(token: str, usr_token: bool) -> str:
    _token = token or os.environ.get("FEDN_AUTH_TOKEN", None)

    if _token is None:
        context_path = os.path.join(home_dir, ".fedn")
        try:
            context_data = get_context(context_path)
            if usr_token:
                _token = context_data.get("User tokens").get("access")
            else:
                _token = context_data.get("Active project tokens").get("access")
        except Exception as e:
            click.secho(f"Encountered error {e}. Make sure you are logged in and have activated a project.", fg="red")

    scheme = os.environ.get("FEDN_AUTH_SCHEME", "Bearer")

    return f"{scheme} {_token}"


def get_client_package_dir(path: str) -> str:
    return path or os.environ.get("FEDN_PACKAGE_DIR", None)


# Print response from api (list of entities)
def print_response(response, entity_name: str, so):
    """Prints the api response to the cli.
    :param response:
        type: array
        description: list of entities
    :param entity_name:
        type: string
        description: name of entity
    :param so:
        type: boolean
        desriptions: single output format
    return: None
    """
    if response.status_code == 200:
        json_data = response.json()
        if so:
            click.echo(f"Found {entity_name}")
            click.echo("\n---------------------------------\n")
            for k, v in json_data.items():
                click.echo(f"\t{k}: {v}")
        else:
            count, result = json_data.values()
            click.echo(f"Found {count} {entity_name}")
            click.echo("\n---------------------------------\n")
            for obj in result:
                print(obj.get("session_id"))
                click.echo("{")
                for k, v in obj.items():
                    click.echo(f"\t{k}: {v}")
                click.echo("}")
    elif response.status_code == 500:
        json_data = response.json()
        click.echo(f'Error: {json_data["message"]}')
    else:
        click.echo(f"Error: {response.status_code}")


def set_context(context_path, context_data):
    """Saves context data as yaml file in given path"""
    try:
        with open(f"{context_path}/context.yaml", "w") as yaml_file:
            yaml.dump(context_data, yaml_file, default_flow_style=False)
    except Exception as e:
        print(f"Error: Failed to write to YAML file. Details: {e}")


def get_context(context_path):
    """Retrieves context data from yaml file in given path"""
    try:
        with open(f"{context_path}/context.yaml", "r") as yaml_file:
            context_data = yaml.safe_load(yaml_file)
    except Exception as e:
        print(f"Error: Failed to write to YAML file. Details: {e}")
    return context_data


def get_response(protocol: str, host: str, port: str, endpoint: str, token: str, headers: dict, usr_api: bool, usr_token: str):
    """Utility function to retrieve response from get request based on provided information."""
    url = get_api_url(protocol=protocol, host=host, port=port, endpoint=endpoint, usr_api=usr_api)

    _token = get_token(token=token, usr_token=usr_token)

    if _token:
        headers["Authorization"] = _token

    response = requests.get(url, headers=headers)

    return response



import docker
import time
import os

CONTAINER_NAME="mongo-test-db"
""" 
    An alternative to this is to create the MongoDB container using
    githiub action and then use the container to create the client
"""
def start_mongodb_container():
    client = docker.from_env()        
    try:
        container = client.containers.get(CONTAINER_NAME)
    except docker.errors.NotFound:
        container = client.containers.run(
            "mongo:7.0",
            detach=True,
            ports={"27017/tcp": None}, # Let Docker choose an available port
            name=CONTAINER_NAME,
            environment={
                "MONGO_INITDB_ROOT_USERNAME": os.environ.get("UNITTEST_DBUSER", "_"),
                "MONGO_INITDB_ROOT_PASSWORD": os.environ.get("UNITTEST_DBPASS", "_"),
            },
            command="mongod"
        )
        
    time.sleep(1)
    start = time.time()
    while time.time() - start < 10:
        try:
            container.reload()
            port = int(container.attrs['NetworkSettings']['Ports']['27017/tcp'][0]['HostPort'])
            break
        except:
            time.sleep(1)
    else:
        raise Exception("Could not start MongoDB container")

    return container, port

def stop_mongodb_container():
    client = docker.from_env()
    container = client.containers.get(CONTAINER_NAME)
    container.stop()
    container.remove(v=True, force=True)
    try:
        container.wait(condition="removed")
    except docker.errors.NotFound:
        pass





import uuid

import click

from fedn.cli.main import main
from fedn.cli.shared import CONTROLLER_DEFAULTS, apply_config, get_response, print_response
from fedn.common.exceptions import InvalidClientConfig
from fedn.network.clients.client_v2 import Client as ClientV2
from fedn.network.clients.client_v2 import ClientOptions


def validate_client_config(config):
    """Validate client configuration.

    :param config: Client config (dict).
    """
    try:
        if config["discover_host"] is None or config["discover_host"] == "":
            if config["combiner"] is None or config["combiner"] == "":
                raise InvalidClientConfig("Missing required configuration: discover_host or combiner")
        if "discover_port" not in config.keys():
            config["discover_port"] = None
        if config["remote_compute_context"] and config["discover_host"] is None:
            raise InvalidClientConfig("Remote compute context requires discover_host")
    except Exception:
        raise InvalidClientConfig("Could not load config from file. Check config")


@main.group("client")
@click.pass_context
def client_cmd(ctx):
    """- Commands for listing and running clients."""
    pass


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("--n_max", required=False, help="Number of items to list")
@client_cmd.command("list")
@click.pass_context
def list_clients(ctx, protocol: str, host: str, port: str, token: str = None, n_max: int = None):
    """Return:
    ------
    - count: number of clients
    - result: list of clients

    """
    headers = {}

    if n_max:
        headers["X-Limit"] = n_max

    response = get_response(protocol=protocol, host=host, port=port, endpoint="clients", token=token, headers=headers, usr_api=False, usr_token=False)
    print_response(response, "clients", None)


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("-id", "--id", required=True, help="Client ID")
@client_cmd.command("get")
@click.pass_context
def get_client(ctx, protocol: str, host: str, port: str, token: str = None, id: str = None):
    """Return:
    ------
    - result: client with given id

    """
    response = get_response(protocol=protocol, host=host, port=port, endpoint=f"clients/{id}", token=token, headers={}, usr_api=False, usr_token=False)
    print_response(response, "client", id)


def _validate_client_params(config: dict):
    api_url = config["api_url"]
    combiner = config["combiner"]
    combiner_port = config["combiner_port"]
    if (api_url is None or api_url == "") and (combiner is None or combiner == ""):
        click.echo("Error: Missing required parameter: --api_url or --combiner")
        return False
    if (combiner is not None and combiner != "") and (combiner_port is None or combiner_port == ""):
        click.echo("Error: Missing required parameter: --combiner-port")
        return False
    return True


def _complement_client_params(config: dict):
    api_url = config["api_url"]
    if not api_url.startswith("http://") and not api_url.startswith("https://"):
        if "localhost" in api_url or "127.0.0.1" in api_url:
            config["api_url"] = "http://" + api_url
        else:
            config["api_url"] = "https://" + api_url
        result = config["api_url"]

        click.echo(f"Protocol missing, complementing api_url with protocol: {result}")


@client_cmd.command("start")
@click.option("-u", "--api-url", required=False, help="Hostname for fedn api.")
@click.option("-p", "--api-port", required=False, help="Port for discovery services (reducer).")
@click.option("--token", required=False, help="Set token provided by reducer if enabled")
@click.option("-n", "--name", required=False)
@click.option("-i", "--client-id", required=False)
@click.option("--local-package", is_flag=True, help="Enable local compute package")
@click.option("-c", "--preferred-combiner", type=str, required=False, default="", help="name of the preferred combiner")
@click.option("--combiner", type=str, required=False, default=None, help="Skip combiner assignment from discover service and attach directly to combiner host.")
@click.option("--combiner-port", type=str, required=False, default=None, help="Combiner port, need to be used with --combiner")
@click.option("-va", "--validator", required=False, default=None)
@click.option("-tr", "--trainer", required=False, default=None)
@click.option("-h", "--helper_type", required=False, default=None)
@click.option("-in", "--init", required=False, default=None, help="Set to a filename to (re)init client from file state.")
@click.pass_context
def client_start_v2_cmd(
    ctx,
    api_url: str,
    api_port: int,
    token: str,
    name: str,
    client_id: str,
    local_package: bool,
    preferred_combiner: str,
    combiner: str,
    combiner_port: int,
    validator: bool,
    trainer: bool,
    helper_type: str,
    init: str,
):
    package = "local" if local_package else "remote"

    config = {
        "api_url": None,
        "api_port": None,
        "token": None,
        "name": None,
        "client_id": None,
        "preferred_combiner": None,
        "combiner": None,
        "combiner_port": None,
        "validator": None,
        "trainer": None,
        "package_checksum": None,
        "helper_type": None,
        # to cater for old inputs
        "discover_host": None,
        "discover_port": None,
    }

    if init:
        apply_config(init, config)
        click.echo(f"\nClient configuration loaded from file: {init}")

        # to cater for old inputs
        if config["discover_host"] is not None:
            config["api_url"] = config["discover_host"]

        if config["discover_port"] is not None:
            config["api_port"] = config["discover_port"]

    if api_url and api_url != "":
        config["api_url"] = api_url
        if config["api_url"] and config["api_url"] != "":
            click.echo(f"Input param api_url: {api_url} overrides value from file")

    if api_port:
        config["api_port"] = api_port
        if config["api_port"]:
            click.echo(f"Input param api_port: {api_port} overrides value from file")

    if token and token != "":
        config["token"] = token
        if config["token"]:
            click.echo(f"Input param token: {token} overrides value from file")

    if name and name != "":
        config["name"] = name
        if config["name"]:
            click.echo(f"Input param name: {name} overrides value from file")
    elif config["name"] is None:
        config["name"] = "client" + str(uuid.uuid4())[:8]

    if client_id and client_id != "":
        config["client_id"] = client_id
        if config["client_id"]:
            click.echo(f"Input param client_id: {client_id} overrides value from file")

    if preferred_combiner and preferred_combiner != "":
        config["preferred_combiner"] = preferred_combiner
        if config["preferred_combiner"]:
            click.echo(f"Input param preferred_combiner: {preferred_combiner} overrides value from file")

    if combiner and combiner != "":
        config["combiner"] = combiner
        if config["combiner"]:
            click.echo(f"Input param combiner: {combiner} overrides value from file")

    if combiner_port:
        config["combiner_port"] = combiner_port
        if config["combiner_port"]:
            click.echo(f"Input param combiner_port: {combiner_port} overrides value from file")

    if validator is not None:
        config["validator"] = validator
        if config["validator"] is not None:
            click.echo(f"Input param validator: {validator} overrides value from file")
    elif config["validator"] is None:
        config["validator"] = True

    if trainer is not None:
        config["trainer"] = trainer
        if config["trainer"] is not None:
            click.echo(f"Input param trainer: {trainer} overrides value from file")
    elif config["trainer"] is None:
        config["trainer"] = True

    if helper_type and helper_type != "":
        config["helper_type"] = helper_type
        if config["helper_type"]:
            click.echo(f"Input param helper_type: {helper_type} overrides value from file")

    if not _validate_client_params(config):
        return

    if config["api_url"] is not None and config["api_url"] != "":
        _complement_client_params(config)

    client_options = ClientOptions(
        name=config["name"],
        package=package,
        preferred_combiner=config["preferred_combiner"],
        id=config["client_id"],
    )
    client = ClientV2(
        api_url=config["api_url"],
        api_port=config["api_port"],
        client_obj=client_options,
        combiner_host=config["combiner"],
        combiner_port=config["combiner_port"],
        token=config["token"],
        package_checksum=config["package_checksum"],
        helper_type=config["helper_type"],
    )

    client.start()



import click

from .main import main
from .shared import CONTROLLER_DEFAULTS, get_response, print_response


@main.group("session")
@click.pass_context
def session_cmd(ctx):
    """:param ctx:"""
    pass


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("--n_max", required=False, help="Number of items to list")
@session_cmd.command("list")
@click.pass_context
def list_sessions(ctx, protocol: str, host: str, port: str, token: str = None, n_max: int = None):
    """Return:
    ------
    - count: number of sessions
    - result: list of sessions

    """
    headers = {}

    if n_max:
        headers["X-Limit"] = n_max

    response = get_response(protocol=protocol, host=host, port=port, endpoint="sessions/", token=token, headers=headers, usr_api=False, usr_token=False)
    print_response(response, "sessions", None)


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("-id", "--id", required=True, help="Session ID")
@session_cmd.command("get")
@click.pass_context
def get_session(ctx, protocol: str, host: str, port: str, token: str = None, id: str = None):
    """Return:
    ------
    - result: session with given session id

    """
    response = get_response(protocol=protocol, host=host, port=port, endpoint=f"sessions/{id}", token=token, headers={}, usr_api=False, usr_token=False)
    print_response(response, "session", id)
